@@ -96,17 +96,20 @@ void MarkerHandleSet::RemoveHandle(int handle) {
 	}
 }
 
-void MarkerHandleSet::RemoveNumber(int markerNum) {
+bool MarkerHandleSet::RemoveNumber(int markerNum) {
+	bool performedDeletion = false;
 	MarkerHandleNumber **pmhn = &root;
 	while (*pmhn) {
 		MarkerHandleNumber *mhn = *pmhn;
 		if (mhn->number == markerNum) {
 			*pmhn = mhn->next;
 			delete mhn;
-			return ;
+			performedDeletion = true;
+		} else {
+			pmhn = &((*pmhn)->next);
 		}
-		pmhn = &((*pmhn)->next);
 	}
+	return performedDeletion;
 }
 
 void MarkerHandleSet::CombineWith(MarkerHandleSet *other) {
@@ -125,6 +128,7 @@ LineVector::LineVector() {
 	levels = 0;
 	sizeLevels = 0;
 	handleCurrent = 1;
+	growSize = 1000;
 
 	Init();
 }
@@ -199,6 +203,8 @@ void LineVector::ClearLevels() {
 void LineVector::InsertValue(int pos, int value) {
 	//Platform::DebugPrintf("InsertValue[%d] = %d\n", pos, value);
 	if ((lines + 2) >= size) {
+		if (growSize * 6 < size)
+			growSize *= 2;
 		Expand(size + growSize);
 		if (levels) {
 			ExpandLevels(size + growSize);
@@ -303,13 +309,18 @@ void LineVector::MergeMarkers(int pos) {
 	}
 }
 
-void LineVector::DeleteMark(int line, int markerNum) {
+void LineVector::DeleteMark(int line, int markerNum, bool all) {
 	if (linesData[line].handleSet) {
 		if (markerNum == -1) {
 			delete linesData[line].handleSet;
 			linesData[line].handleSet = 0;
 		} else {
-			linesData[line].handleSet->RemoveNumber(markerNum);
+			bool performedDeletion = 
+				linesData[line].handleSet->RemoveNumber(markerNum);
+			while (all && performedDeletion) {
+				performedDeletion = 
+					linesData[line].handleSet->RemoveNumber(markerNum);
+			}
 			if (linesData[line].handleSet->Length() == 0) {
 				delete linesData[line].handleSet;
 				linesData[line].handleSet = 0;
@@ -439,6 +450,9 @@ void UndoHistory::AppendAction(actionType at, int position, char *data, int leng
 	//Platform::DebugPrintf("%% %d action %d %d %d\n", at, position, lengthData, currentAction);
 	//Platform::DebugPrintf("^ %d action %d %d\n", actions[currentAction - 1].at,
 	//	actions[currentAction - 1].position, actions[currentAction - 1].lenData);
+	if (currentAction < savePoint) {
+		savePoint = -1;
+	}
 	if (currentAction >= 1) {
 		if (0 == undoSequenceDepth) {
 			// Top level actions may not always be coalesced
@@ -624,21 +638,11 @@ void CellBuffer::RoomFor(int insertionLength) {
 	//Platform::DebugPrintf("need room %d %d\n", gaplen, insertionLength);
 	if (gaplen <= insertionLength) {
 		//Platform::DebugPrintf("need room %d %d\n", gaplen, insertionLength);
-		GapTo(length);
 		if (growSize * 6 < size)
 			growSize *= 2;
 		int newSize = size + insertionLength + growSize;
-		//Platform::DebugPrintf("moved gap %d\n", newSize);
-		char *newBody = new char[newSize];
-		memcpy(newBody, body, size);
-		delete []body;
-		body = newBody;
-		gaplen += newSize - size;
-		part2body = body + gaplen;
-		size = newSize;
-		//Platform::DebugPrintf("end need room %d %d - size=%d length=%d\n", gaplen, insertionLength,size,length);
+		Allocate(newSize);
 	}
-
 }
 
 // To make it easier to write code that uses ByteAt, a position outside the range of the buffer
@@ -736,6 +740,7 @@ void CellBuffer::InsertCharStyle(int position, char ch, char style) {
 }
 
 bool CellBuffer::SetStyleAt(int position, char style, char mask) {
+	style &= mask;
 	char curVal = ByteAt(position * 2 + 1);
 	if ((curVal & mask) != style) {
 		SetByteAt(position*2 + 1, static_cast<char>((curVal & ~mask) | style));
@@ -787,6 +792,19 @@ int CellBuffer::Length() {
 	return ByteLength() / 2;
 }
 
+void CellBuffer::Allocate(int newSize) {
+	if (newSize > length) {
+		GapTo(length);
+		char *newBody = new char[newSize];
+		memcpy(newBody, body, length);
+		delete []body;
+		body = newBody;
+		gaplen += newSize - size;
+		part2body = body + gaplen;
+		size = newSize;
+	}
+}
+
 int CellBuffer::Lines() {
 	//Platform::DebugPrintf("Lines = %d\n", lv.lines);
 	return lv.lines;
@@ -826,7 +844,7 @@ int CellBuffer::AddMark(int line, int markerNum) {
 
 void CellBuffer::DeleteMark(int line, int markerNum) {
 	if ((line >= 0) && (line < lv.lines)) {
-		lv.DeleteMark(line, markerNum);
+		lv.DeleteMark(line, markerNum, false);
 	}
 }
 
@@ -842,7 +860,7 @@ int CellBuffer::GetMark(int line) {
 
 void CellBuffer::DeleteAllMarks(int markerNum) {
 	for (int line = 0; line < lv.lines; line++) {
-		lv.DeleteMark(line, markerNum);
+		lv.DeleteMark(line, markerNum, true);
 	}
 }
 
@@ -958,10 +976,13 @@ void CellBuffer::BasicDeleteChars(int position, int deleteLength) {
 					//Platform::DebugPrintf("Removing cr end of line\n");
 					lv.Remove(lineRemove);
 				}
-			} else if ((ch == '\n') && !ignoreNL) {
-				//Platform::DebugPrintf("Removing lf end of line\n");
-				lv.Remove(lineRemove);
-				ignoreNL = false; 	// Further \n are not real deletions
+			} else if (ch == '\n') {
+				if (ignoreNL) {
+					ignoreNL = false; 	// Further \n are real deletions
+				} else {
+					//Platform::DebugPrintf("Removing lf end of line\n");
+					lv.Remove(lineRemove);
+				}
 			}
 
 			ch = chNext;
