@@ -20,6 +20,19 @@
 #include "Scintilla.h"
 #include "SciLexer.h"
 
+static bool Is0To9(char ch) {
+	return (ch >= '0') && (ch <= '9');
+}
+
+static bool Is1To9(char ch) {
+	return (ch >= '1') && (ch <= '9');
+}
+
+static inline bool AtEOL(Accessor &styler, unsigned int i) {
+	return (styler[i] == '\n') ||
+		((styler[i] == '\r') && (styler.SafeGetCharAt(i + 1) != '\n'));
+}
+
 static void ColouriseBatchLine(
     char *lineBuffer,
     unsigned int lengthLine,
@@ -31,13 +44,13 @@ static void ColouriseBatchLine(
 	unsigned int i = 0;
 	unsigned int state = SCE_BAT_DEFAULT;
 
-	while (isspacechar(lineBuffer[i]) && (i < lengthLine)) {	// Skip initial spaces
+	while ((i < lengthLine) && isspacechar(lineBuffer[i])) {	// Skip initial spaces
 		i++;
 	}
 	if (lineBuffer[i] == '@') {	// Hide command (ECHO OFF)
 		styler.ColourTo(startLine + i, SCE_BAT_HIDE);
 		i++;
-		while (isspacechar(lineBuffer[i]) && (i < lengthLine)) {	// Skip next spaces
+		while ((i < lengthLine) && isspacechar(lineBuffer[i])) {	// Skip next spaces
 			i++;
 		}
 	}
@@ -80,7 +93,7 @@ static void ColouriseBatchLine(
 		while (offset < lengthLine) {
 			if (state == SCE_BAT_DEFAULT && lineBuffer[offset] == '%') {
 				styler.ColourTo(startLine + offset - 1, state);
-				if (isdigit(lineBuffer[offset + 1])) {
+				if (Is0To9(lineBuffer[offset + 1])) {
 					styler.ColourTo(startLine + offset + 1, SCE_BAT_IDENTIFIER);
 					offset += 2;
 				} else if (lineBuffer[offset + 1] == '%' &&
@@ -130,23 +143,20 @@ static void ColouriseBatchDoc(
 
 	styler.StartAt(startPos);
 	styler.StartSegment(startPos);
-	unsigned int linePos = 0, startLine = startPos;
-	for (unsigned int i = startPos; i <= startPos + length; i++) {
+	unsigned int linePos = 0;
+	unsigned int startLine = startPos;
+	for (unsigned int i = startPos; i < startPos + length; i++) {
 		lineBuffer[linePos++] = styler[i];
-		if (styler[i] == '\r' || styler[i] == '\n' || (linePos >=
-		        sizeof(lineBuffer) - 1)) {
+		if (AtEOL(styler, i) || (linePos >= sizeof(lineBuffer) - 1)) {
 			// End of line (or of line buffer) met, colourise it
-			if (styler[i + 1] == '\n') {
-				lineBuffer[linePos++] = styler[++i];
-			}
 			lineBuffer[linePos] = '\0';
 			ColouriseBatchLine(lineBuffer, linePos, startLine, i, keywords, styler);
 			linePos = 0;
 			startLine = i + 1;
 		}
 	}
-	if (linePos > 0) {
-		ColouriseBatchLine(lineBuffer, linePos, startLine, startPos + length,
+	if (linePos > 0) {	// Last line does not have ending characters
+		ColouriseBatchLine(lineBuffer, linePos, startLine, startPos + length - 1,
 		                   keywords, styler);
 	}
 }
@@ -156,22 +166,47 @@ static void ColouriseDiffLine(char *lineBuffer, int endLine, Accessor &styler) {
 	// comment lines before the first "diff " or "--- ". If a real
 	// difference starts then each line starting with ' ' is a whitespace
 	// otherwise it is considered a comment (Only in..., Binary file...)
-	if (0 == strncmp(lineBuffer, "diff ", 3)) {
-		styler.ColourTo(endLine, 2);
-	} else if (0 == strncmp(lineBuffer, "--- ", 3)) {
-		styler.ColourTo(endLine, 3);
-	} else if (0 == strncmp(lineBuffer, "+++ ", 3)) {
-		styler.ColourTo(endLine, 3);
+	if (0 == strncmp(lineBuffer, "diff ", 5)) {
+		styler.ColourTo(endLine, SCE_DIFF_COMMAND);
+	} else if (0 == strncmp(lineBuffer, "--- ", 4)) {
+		// In a context diff, --- appears in both the header and the position markers
+		if (atoi(lineBuffer+4) && !strchr(lineBuffer, '/'))
+			styler.ColourTo(endLine, SCE_DIFF_POSITION);
+		else
+			styler.ColourTo(endLine, SCE_DIFF_HEADER);
+	} else if (0 == strncmp(lineBuffer, "+++ ", 4)) {
+		// I don't know of any diff where "+++ " is a position marker, but for
+		// consistency, do the same as with "--- " and "*** ".
+		if (atoi(lineBuffer+4) && !strchr(lineBuffer, '/'))
+			styler.ColourTo(endLine, SCE_DIFF_POSITION);
+		else
+			styler.ColourTo(endLine, SCE_DIFF_HEADER);
+	} else if (0 == strncmp(lineBuffer, "====", 4)) {  // For p4's diff
+		styler.ColourTo(endLine, SCE_DIFF_HEADER);
+	} else if (0 == strncmp(lineBuffer, "***", 3)) {
+		// In a context diff, *** appears in both the header and the position markers.
+		// Also ******** is a chunk header, but here it's treated as part of the
+		// position marker since there is no separate style for a chunk header.
+		if (lineBuffer[3] == ' ' && atoi(lineBuffer+4) && !strchr(lineBuffer, '/'))
+			styler.ColourTo(endLine, SCE_DIFF_POSITION);
+		else if (lineBuffer[3] == '*')
+			styler.ColourTo(endLine, SCE_DIFF_POSITION);
+		else
+			styler.ColourTo(endLine, SCE_DIFF_HEADER);
+	} else if (0 == strncmp(lineBuffer, "? ", 2)) {    // For difflib
+		styler.ColourTo(endLine, SCE_DIFF_HEADER);
 	} else if (lineBuffer[0] == '@') {
-		styler.ColourTo(endLine, 4);
-	} else if (lineBuffer[0] == '-') {
-		styler.ColourTo(endLine, 5);
-	} else if (lineBuffer[0] == '+') {
-		styler.ColourTo(endLine, 6);
+		styler.ColourTo(endLine, SCE_DIFF_POSITION);
+	} else if (lineBuffer[0] >= '0' && lineBuffer[0] <= '9') {
+		styler.ColourTo(endLine, SCE_DIFF_POSITION);
+	} else if (lineBuffer[0] == '-' || lineBuffer[0] == '<') {
+		styler.ColourTo(endLine, SCE_DIFF_DELETED);
+	} else if (lineBuffer[0] == '+' || lineBuffer[0] == '>') {
+		styler.ColourTo(endLine, SCE_DIFF_ADDED);
 	} else if (lineBuffer[0] != ' ') {
-		styler.ColourTo(endLine, 1);
+		styler.ColourTo(endLine, SCE_DIFF_COMMENT);
 	} else {
-		styler.ColourTo(endLine, 0);
+		styler.ColourTo(endLine, SCE_DIFF_DEFAULT);
 	}
 }
 
@@ -182,15 +217,49 @@ static void ColouriseDiffDoc(unsigned int startPos, int length, int, WordList *[
 	unsigned int linePos = 0;
 	for (unsigned int i = startPos; i < startPos + length; i++) {
 		lineBuffer[linePos++] = styler[i];
-		if (styler[i] == '\r' || styler[i] == '\n' || (linePos >= sizeof(lineBuffer) - 1)) {
+		if (AtEOL(styler, i) || (linePos >= sizeof(lineBuffer) - 1)) {
+			// End of line (or of line buffer) met, colourise it
+			lineBuffer[linePos] = '\0';
 			ColouriseDiffLine(lineBuffer, i, styler);
 			linePos = 0;
 		}
 	}
-	if (linePos > 0)
-		ColouriseDiffLine(lineBuffer, startPos + length, styler);
+	if (linePos > 0) {	// Last line does not have ending characters
+		ColouriseDiffLine(lineBuffer, startPos + length - 1, styler);
+	}
 }
 
+static void FoldDiffDoc(unsigned int startPos, int length, int, WordList*[], Accessor &styler) {
+	int curLine = styler.GetLine(startPos);
+	int prevLevel = SC_FOLDLEVELBASE;
+	if (curLine > 0)
+		prevLevel = styler.LevelAt(curLine-1);
+
+	int curLineStart = styler.LineStart(curLine);
+	do {
+		int nextLevel = prevLevel;
+		if (prevLevel & SC_FOLDLEVELHEADERFLAG)
+			nextLevel = (prevLevel & SC_FOLDLEVELNUMBERMASK) + 1;
+		
+		int lineType = styler.StyleAt(curLineStart);
+		if (lineType == SCE_DIFF_COMMAND)
+			nextLevel = (SC_FOLDLEVELBASE + 1) | SC_FOLDLEVELHEADERFLAG;
+		else if (lineType == SCE_DIFF_HEADER) {
+			nextLevel = (SC_FOLDLEVELBASE + 2) | SC_FOLDLEVELHEADERFLAG;
+		} else if (lineType == SCE_DIFF_POSITION)
+			nextLevel = (SC_FOLDLEVELBASE + 3) | SC_FOLDLEVELHEADERFLAG;
+		
+		if ((nextLevel & SC_FOLDLEVELHEADERFLAG) && (nextLevel == prevLevel))
+			styler.SetLevel(curLine-1, prevLevel & ~SC_FOLDLEVELHEADERFLAG);
+
+		styler.SetLevel(curLine, nextLevel);
+		prevLevel = nextLevel;
+		
+		curLineStart = styler.LineStart(++curLine);
+	} while (static_cast<int>(startPos) + length > curLineStart);
+}
+
+
 static void ColourisePropsLine(
     char *lineBuffer,
     unsigned int lengthLine,
@@ -199,28 +268,32 @@ static void ColourisePropsLine(
     Accessor &styler) {
 
 	unsigned int i = 0;
-	while (isspacechar(lineBuffer[i]) && (i < lengthLine))	// Skip initial spaces
+	while ((i < lengthLine) && isspacechar(lineBuffer[i]))	// Skip initial spaces
 		i++;
-	if (lineBuffer[i] == '#' || lineBuffer[i] == '!' || lineBuffer[i] == ';') {
-		styler.ColourTo(endPos, 1);
-	} else if (lineBuffer[i] == '[') {
-		styler.ColourTo(endPos, 2);
-	} else if (lineBuffer[i] == '@') {
-		styler.ColourTo(startLine + i, 4);
-		if (lineBuffer[++i] == '=')
-			styler.ColourTo(startLine + i, 3);
-		styler.ColourTo(endPos, 0);
-	} else {
-		// Search for the '=' character
-		while (lineBuffer[i] != '=' && (i < lengthLine - 1))
-			i++;
-		if (lineBuffer[i] == '=') {
-			styler.ColourTo(startLine + i - 1, 0);
-			styler.ColourTo(startLine + i, 3);
-			styler.ColourTo(endPos, 0);
+	if (i < lengthLine) {
+		if (lineBuffer[i] == '#' || lineBuffer[i] == '!' || lineBuffer[i] == ';') {
+			styler.ColourTo(endPos, SCE_PROPS_COMMENT);
+		} else if (lineBuffer[i] == '[') {
+			styler.ColourTo(endPos, SCE_PROPS_SECTION);
+		} else if (lineBuffer[i] == '@') {
+			styler.ColourTo(startLine + i, SCE_PROPS_DEFVAL);
+			if (lineBuffer[++i] == '=')
+				styler.ColourTo(startLine + i, SCE_PROPS_ASSIGNMENT);
+			styler.ColourTo(endPos, SCE_PROPS_DEFAULT);
 		} else {
-			styler.ColourTo(endPos, 0);
+			// Search for the '=' character
+			while ((i < lengthLine) && (lineBuffer[i] != '='))
+				i++;
+			if ((i < lengthLine) && (lineBuffer[i] == '=')) {
+				styler.ColourTo(startLine + i - 1, SCE_PROPS_DEFAULT);
+				styler.ColourTo(startLine + i, 3);
+				styler.ColourTo(endPos, SCE_PROPS_DEFAULT);
+			} else {
+				styler.ColourTo(endPos, SCE_PROPS_DEFAULT);
+			}
 		}
+	} else {
+		styler.ColourTo(endPos, SCE_PROPS_DEFAULT);
 	}
 }
 
@@ -228,20 +301,95 @@ static void ColourisePropsDoc(unsigned int startPos, int length, int, WordList *
 	char lineBuffer[1024];
 	styler.StartAt(startPos);
 	styler.StartSegment(startPos);
-	unsigned int linePos = 0, startLine = startPos;
-	for (unsigned int i = startPos; i <= startPos + length; i++) {
+	unsigned int linePos = 0;
+	unsigned int startLine = startPos;
+	for (unsigned int i = startPos; i < startPos + length; i++) {
 		lineBuffer[linePos++] = styler[i];
-		if ((styler[i] == '\r' && styler.SafeGetCharAt(i + 1) != '\n') ||
-		        styler[i] == '\n' ||
-		        (linePos >= sizeof(lineBuffer) - 1)) {
+		if (AtEOL(styler, i) || (linePos >= sizeof(lineBuffer) - 1)) {
+			// End of line (or of line buffer) met, colourise it
 			lineBuffer[linePos] = '\0';
 			ColourisePropsLine(lineBuffer, linePos, startLine, i, styler);
 			linePos = 0;
 			startLine = i + 1;
 		}
 	}
-	if (linePos > 0)
-		ColourisePropsLine(lineBuffer, linePos, startLine, startPos + length, styler);
+	if (linePos > 0) {	// Last line does not have ending characters
+		ColourisePropsLine(lineBuffer, linePos, startLine, startPos + length - 1, styler);
+	}
+}
+
+// adaption by ksc, using the "} else {" trick of 1.53
+// 030721
+static void FoldPropsDoc(unsigned int startPos, int length, int, WordList *[], Accessor &styler) {
+	bool foldCompact = styler.GetPropertyInt("fold.compact", 1) != 0;
+
+	unsigned int endPos = startPos + length;
+	int visibleChars = 0;
+	int lineCurrent = styler.GetLine(startPos);
+
+	char chNext = styler[startPos];
+	int styleNext = styler.StyleAt(startPos);
+	bool headerPoint = false;
+	int lev;
+
+	for (unsigned int i = startPos; i < endPos; i++) {
+		char ch = chNext;
+		chNext = styler[i+1];
+
+		int style = styleNext;
+		styleNext = styler.StyleAt(i + 1);
+		bool atEOL = (ch == '\r' && chNext != '\n') || (ch == '\n');
+
+		if (style == SCE_PROPS_SECTION) {
+			headerPoint = true;
+		}
+
+		if (atEOL) {
+			lev = SC_FOLDLEVELBASE;
+
+			if (lineCurrent > 0) {
+				int levelPrevious = styler.LevelAt(lineCurrent - 1);
+
+				if (levelPrevious & SC_FOLDLEVELHEADERFLAG) {
+					lev = SC_FOLDLEVELBASE + 1;
+				} else {
+					lev = levelPrevious & SC_FOLDLEVELNUMBERMASK;
+				}
+			}
+
+			if (headerPoint) {
+				lev = SC_FOLDLEVELBASE;
+			}
+			if (visibleChars == 0 && foldCompact)
+				lev |= SC_FOLDLEVELWHITEFLAG;
+
+			if (headerPoint) {
+				lev |= SC_FOLDLEVELHEADERFLAG;
+			}
+			if (lev != styler.LevelAt(lineCurrent)) {
+				styler.SetLevel(lineCurrent, lev);
+			}
+
+			lineCurrent++;
+			visibleChars = 0;
+			headerPoint=false;
+		}
+		if (!isspacechar(ch))
+			visibleChars++;
+	}
+
+	if (lineCurrent > 0) {
+		int levelPrevious = styler.LevelAt(lineCurrent - 1);
+		if (levelPrevious & SC_FOLDLEVELHEADERFLAG) {
+			lev = SC_FOLDLEVELBASE + 1;
+		} else {
+			lev = levelPrevious & SC_FOLDLEVELNUMBERMASK;
+		}
+	} else {
+		lev = SC_FOLDLEVELBASE;
+	}
+	int flagsNext = styler.LevelAt(lineCurrent);
+	styler.SetLevel(lineCurrent, lev | flagsNext & ~SC_FOLDLEVELNUMBERMASK);
 }
 
 static void ColouriseMakeLine(
@@ -252,11 +400,11 @@ static void ColouriseMakeLine(
     Accessor &styler) {
 
 	unsigned int i = 0;
-        unsigned int lastNonSpace = 0;
+	int lastNonSpace = -1;
 	unsigned int state = SCE_MAKE_DEFAULT;
 	bool bSpecial = false;
 	// Skip initial spaces
-	while (isspacechar(lineBuffer[i]) && (i < lengthLine)) {
+	while ((i < lengthLine) && isspacechar(lineBuffer[i])) {
 		i++;
 	}
 	if (lineBuffer[i] == '#') {	// Comment
@@ -277,13 +425,17 @@ static void ColouriseMakeLine(
 		}
 		if (!bSpecial) {
 			if (lineBuffer[i] == ':') {
-				styler.ColourTo(startLine + lastNonSpace, SCE_MAKE_TARGET);
+				// We should check that no colouring was made since the beginning of the line,
+				// to avoid colouring stuff like /OUT:file
+				if (lastNonSpace >= 0)
+					styler.ColourTo(startLine + lastNonSpace, SCE_MAKE_TARGET);
 				styler.ColourTo(startLine + i - 1, SCE_MAKE_DEFAULT);
 				styler.ColourTo(startLine + i, SCE_MAKE_OPERATOR);
 				bSpecial = true;	// Only react to the first ':' of the line
 				state = SCE_MAKE_DEFAULT;
 			} else if (lineBuffer[i] == '=') {
-				styler.ColourTo(startLine + lastNonSpace, SCE_MAKE_IDENTIFIER);
+				if (lastNonSpace >= 0)
+					styler.ColourTo(startLine + lastNonSpace, SCE_MAKE_IDENTIFIER);
 				styler.ColourTo(startLine + i - 1, SCE_MAKE_DEFAULT);
 				styler.ColourTo(startLine + i, SCE_MAKE_OPERATOR);
 				bSpecial = true;	// Only react to the first '=' of the line
@@ -306,30 +458,40 @@ static void ColouriseMakeDoc(unsigned int startPos, int length, int, WordList *[
 	char lineBuffer[1024];
 	styler.StartAt(startPos);
 	styler.StartSegment(startPos);
-	unsigned int linePos = 0, startLine = startPos;
-	for (unsigned int i = startPos; i <= startPos + length; i++) {
+	unsigned int linePos = 0;
+	unsigned int startLine = startPos;
+	for (unsigned int i = startPos; i < startPos + length; i++) {
 		lineBuffer[linePos++] = styler[i];
-		if (styler[i] == '\r' || styler[i] == '\n' || (linePos >= sizeof(lineBuffer) - 1)) {
+		if (AtEOL(styler, i) || (linePos >= sizeof(lineBuffer) - 1)) {
+			// End of line (or of line buffer) met, colourise it
 			lineBuffer[linePos] = '\0';
 			ColouriseMakeLine(lineBuffer, linePos, startLine, i, styler);
 			linePos = 0;
 			startLine = i + 1;
 		}
 	}
-	if (linePos > 0) {
-		ColouriseMakeLine(lineBuffer, linePos, startLine, startPos + length, styler);
+	if (linePos > 0) {	// Last line does not have ending characters
+		ColouriseMakeLine(lineBuffer, linePos, startLine, startPos + length - 1, styler);
 	}
 }
 
+static bool strstart(char *haystack, char *needle) {
+	return strncmp(haystack, needle, strlen(needle)) == 0;
+}
+
 static void ColouriseErrorListLine(
     char *lineBuffer,
     unsigned int lengthLine,
     //		unsigned int startLine,
     unsigned int endPos,
     Accessor &styler) {
+	const int unRecognized = 99;
 	if (lineBuffer[0] == '>') {
 		// Command or return status
 		styler.ColourTo(endPos, SCE_ERR_CMD);
+	} else if (lineBuffer[0] == '<') {
+		// Diff removal, but not interested. Trapped to avoid hitting CTAG cases.
+		styler.ColourTo(endPos, SCE_ERR_DEFAULT);
 	} else if (lineBuffer[0] == '!') {
 		styler.ColourTo(endPos, SCE_ERR_DIFF_CHANGED);
 	} else if (lineBuffer[0] == '+') {
@@ -338,66 +500,139 @@ static void ColouriseErrorListLine(
 		styler.ColourTo(endPos, SCE_ERR_DIFF_MESSAGE);
 	} else if (lineBuffer[0] == '-') {
 		styler.ColourTo(endPos, SCE_ERR_DIFF_DELETION);
+	} else if (strstart(lineBuffer, "cf90-")) {
+		// Absoft Pro Fortran 90/95 v8.2 error and/or warning message
+		styler.ColourTo(endPos, SCE_ERR_ABSF);
+	} else if (strstart(lineBuffer, "fortcom:")) {
+		// Intel Fortran Compiler v8.0 error/warning message
+		styler.ColourTo(endPos, SCE_ERR_IFORT);
 	} else if (strstr(lineBuffer, "File \"") && strstr(lineBuffer, ", line ")) {
 		styler.ColourTo(endPos, SCE_ERR_PYTHON);
-	} else if (0 == strncmp(lineBuffer, "Error ", strlen("Error "))) {
+	} else if (strstr(lineBuffer, " in ") && strstr(lineBuffer, " on line ")) {
+		styler.ColourTo(endPos, SCE_ERR_PHP);
+	} else if ((strstart(lineBuffer, "Error ") ||
+		strstart(lineBuffer, "Warning ")) &&
+		strstr(lineBuffer, " at (") &&
+		strstr(lineBuffer, ") : ") &&
+		(strstr(lineBuffer, " at (") < strstr(lineBuffer, ") : "))) {
+		// Intel Fortran Compiler error/warning message
+		styler.ColourTo(endPos, SCE_ERR_IFC);
+	} else if (strstart(lineBuffer, "Error ")) {
 		// Borland error message
 		styler.ColourTo(endPos, SCE_ERR_BORLAND);
-	} else if (0 == strncmp(lineBuffer, "Warning ", strlen("Warning "))) {
+	} else if (strstart(lineBuffer, "Warning ")) {
 		// Borland warning message
 		styler.ColourTo(endPos, SCE_ERR_BORLAND);
 	} else if (strstr(lineBuffer, "at line " ) &&
-	           strstr(lineBuffer, "at line " ) < lineBuffer + lengthLine &&
+	           (strstr(lineBuffer, "at line " ) < (lineBuffer + lengthLine)) &&
 	           strstr(lineBuffer, "file ") &&
-	           strstr(lineBuffer, "file ") < lineBuffer + lengthLine) {
-		// Lua error message
+	           (strstr(lineBuffer, "file ") < (lineBuffer + lengthLine))) {
+		// Lua 4 error message
 		styler.ColourTo(endPos, SCE_ERR_LUA);
 	} else if (strstr(lineBuffer, " at " ) &&
-	           strstr(lineBuffer, " at " ) < lineBuffer + lengthLine &&
+	           (strstr(lineBuffer, " at " ) < (lineBuffer + lengthLine)) &&
 	           strstr(lineBuffer, " line ") &&
-	           strstr(lineBuffer, " line ") < lineBuffer + lengthLine) {
+	           (strstr(lineBuffer, " line ") < (lineBuffer + lengthLine)) &&
+			   (strstr(lineBuffer, " at " ) < (strstr(lineBuffer, " line ")))) {
 		// perl error message
 		styler.ColourTo(endPos, SCE_ERR_PERL);
 	} else if ((memcmp(lineBuffer, "   at ", 6) == 0) &&
 		strstr(lineBuffer, ":line ")) {
 		// A .NET traceback
 		styler.ColourTo(endPos, SCE_ERR_NET);
+	} else if (strstart(lineBuffer, "Line ") &&
+		strstr(lineBuffer, ", file ")) {
+		// Essential Lahey Fortran error message
+		styler.ColourTo(endPos, SCE_ERR_ELF);
+	} else if (strstart(lineBuffer, "line ") &&
+		       strstr(lineBuffer, " column ")) {
+		// HTML tidy style: line 42 column 1
+		styler.ColourTo(endPos, SCE_ERR_TIDY);
+	} else if (strstart(lineBuffer, "\tat ") &&
+		strstr(lineBuffer, "(") &&
+		strstr(lineBuffer, ".java:")) {
+		// Java stack back trace
+		styler.ColourTo(endPos, SCE_ERR_JAVA_STACK);
 	} else {
-		// Look for <filename>:<line>:message
-		// Look for <filename>(line)message
-		// Look for <filename>(line,pos)message
+		// Look for GCC <filename>:<line>:message
+		// Look for Microsoft <filename>(line) :message
+		// Look for Microsoft <filename>(line,pos)message
+		// Look for CTags \tmessage
+		// Look for Lua 5 traceback \t<filename>:<line>:message
+		bool initialTab = (lineBuffer[0] == '\t');
 		int state = 0;
 		for (unsigned int i = 0; i < lengthLine; i++) {
-			if (state == 0 && lineBuffer[i] == ':' && isdigit(lineBuffer[i + 1])) {
-				state = 1;
-			} else if (state == 0 && lineBuffer[i] == '(') {
-				state = 10;
-			} else if (state == 1 && isdigit(lineBuffer[i])) {
-				state = 2;
-			} else if (state == 2 && lineBuffer[i] == ':') {
-				state = 3;
+			char ch = lineBuffer[i];
+			char chNext = ' ';
+			if ((i+1) < lengthLine)
+				chNext = lineBuffer[i+1];
+			if (state == 0) {
+				if (ch == ':') {
+					// May be GCC, or might be Lua 5 (Lua traceback same but with tab prefix)
+					if ((chNext != '\\') && (chNext != '/')) {
+						// This check is not completely accurate as may be on
+						// GTK+ with a file name that includes ':'.
+						state = 1;
+					}
+				} else if ((ch == '(') && Is1To9(chNext) && (!initialTab)) {
+					// May be Microsoft
+					// Check against '0' often removes phone numbers
+					state = 10;
+				} else if ((ch == '\t') && (!initialTab)) {
+					// May be CTags
+					state = 20;
+				}
+			} else if (state == 1) {
+				state = Is1To9(ch) ? 2 : unRecognized;
+			} else if (state == 2) {
+				if (ch == ':') {
+					state = 3;	// :9.*: is GCC
+					break;
+				} else if (!Is0To9(ch)) {
+					state = unRecognized;
+				}
+			} else if (state == 10) {
+				state = Is0To9(ch) ? 11 : unRecognized;
+			} else if (state == 11) {
+				if (ch == ',') {
+					state = 14;
+				} else if (ch == ')') {
+					state = 12;
+				} else if ((ch != ' ') && !Is0To9(ch)) {
+					state = unRecognized;
+				}
+			} else if (state == 12) {
+				if ((ch == ' ') && (chNext == ':')) {
+					state = 13;
+				} else {
+					state = unRecognized;
+				}
+			} else if (state == 14) {
+				if (ch == ')') {
+					state = 15;
+					break;
+				} else if ((ch != ' ') && !Is0To9(ch)) {
+					state = unRecognized;
+				}
+			} else if (state == 20) {
+				if ((lineBuffer[i-1] == '\t') &&
+					((ch == '/' && lineBuffer[i+1] == '^') || Is0To9(ch))) {
+					state = 24;
+					break;
+				} else if ((ch == '/') && (lineBuffer[i+1] == '^')) {
+					state = 21;
+				}
+			} else if ((state == 21) && ((lineBuffer[i] == '$') && (lineBuffer[i+1] == '/'))) {
+				state = 22;
 				break;
-			} else if (state == 2 && !isdigit(lineBuffer[i])) {
-				state = 99;
-			} else if (state == 10 && isdigit(lineBuffer[i])) {
-				state = 11;
-			} else if (state == 11 && lineBuffer[i] == ',') {
-				state = 14;
-			} else if (state == 11 && lineBuffer[i] == ')') {
-				state = 12;
-			} else if (state == 12 && lineBuffer[i] == ':') {
-				state = 13;
-			} else if (state == 14 && lineBuffer[i] == ')') {
-				state = 15;
-				break;
-			} else if (((state == 11) || (state == 14)) && !((lineBuffer[i] == ' ') || isdigit(lineBuffer[i]))) {
-				state = 99;
 			}
 		}
 		if (state == 3) {
 			styler.ColourTo(endPos, SCE_ERR_GCC);
 		} else if ((state == 13) || (state == 14) || (state == 15)) {
 			styler.ColourTo(endPos, SCE_ERR_MS);
+		} else if ((state == 22) || (state == 24)) {
+			styler.ColourTo(endPos, SCE_ERR_CTAG);
 		} else {
 			styler.ColourTo(endPos, SCE_ERR_DEFAULT);
 		}
@@ -409,15 +644,18 @@ static void ColouriseErrorListDoc(unsigned int startPos, int length, int, WordLi
 	styler.StartAt(startPos);
 	styler.StartSegment(startPos);
 	unsigned int linePos = 0;
-	for (unsigned int i = startPos; i <= startPos + length; i++) {
+	for (unsigned int i = startPos; i < startPos + length; i++) {
 		lineBuffer[linePos++] = styler[i];
-		if (styler[i] == '\r' || styler[i] == '\n' || (linePos >= sizeof(lineBuffer) - 1)) {
+		if (AtEOL(styler, i) || (linePos >= sizeof(lineBuffer) - 1)) {
+			// End of line (or of line buffer) met, colourise it
+			lineBuffer[linePos] = '\0';
 			ColouriseErrorListLine(lineBuffer, linePos, i, styler);
 			linePos = 0;
 		}
 	}
-	if (linePos > 0)
-		ColouriseErrorListLine(lineBuffer, linePos, startPos + length, styler);
+	if (linePos > 0) {	// Last line does not have ending characters
+		ColouriseErrorListLine(lineBuffer, linePos, startPos + length - 1, styler);
+	}
 }
 
 static int isSpecial(char s) {
@@ -518,12 +756,32 @@ static void ColouriseLatexDoc(unsigned int startPos, int length, int initStyle,
 			}
 		}
 	}
-	styler.ColourTo(lengthDoc, state);
+	styler.ColourTo(lengthDoc-1, state);
 }
 
-LexerModule lmBatch(SCLEX_BATCH, ColouriseBatchDoc, "batch");
-LexerModule lmDiff(SCLEX_DIFF, ColouriseDiffDoc, "diff");
-LexerModule lmProps(SCLEX_PROPERTIES, ColourisePropsDoc, "props");
-LexerModule lmMake(SCLEX_MAKEFILE, ColouriseMakeDoc, "makefile");
-LexerModule lmErrorList(SCLEX_ERRORLIST, ColouriseErrorListDoc, "errorlist");
-LexerModule lmLatex(SCLEX_LATEX, ColouriseLatexDoc, "latex");
+static const char * const batchWordListDesc[] = {
+	"Keywords",
+	0
+};
+
+static const char * const emptyWordListDesc[] = {
+	0
+};
+
+static void ColouriseNullDoc(unsigned int startPos, int length, int, WordList *[],
+                            Accessor &styler) {
+	// Null language means all style bytes are 0 so just mark the end - no need to fill in.
+	if (length > 0) {
+		styler.StartAt(startPos + length - 1);
+		styler.StartSegment(startPos + length - 1);
+		styler.ColourTo(startPos + length - 1, 0);
+	}
+}
+
+LexerModule lmBatch(SCLEX_BATCH, ColouriseBatchDoc, "batch", 0, batchWordListDesc);
+LexerModule lmDiff(SCLEX_DIFF, ColouriseDiffDoc, "diff", FoldDiffDoc, emptyWordListDesc);
+LexerModule lmProps(SCLEX_PROPERTIES, ColourisePropsDoc, "props", FoldPropsDoc, emptyWordListDesc);
+LexerModule lmMake(SCLEX_MAKEFILE, ColouriseMakeDoc, "makefile", 0, emptyWordListDesc);
+LexerModule lmErrorList(SCLEX_ERRORLIST, ColouriseErrorListDoc, "errorlist", 0, emptyWordListDesc);
+LexerModule lmLatex(SCLEX_LATEX, ColouriseLatexDoc, "latex", 0, emptyWordListDesc);
+LexerModule lmNull(SCLEX_NULL, ColouriseNullDoc, "null");
