@@ -2,7 +2,7 @@
 /** @file PlatWin.cxx
  ** Implementation of platform facilities on Windows.
  **/
-// Copyright 1998-2002 by Neil Hodgson <neilh@scintilla.org>
+// Copyright 1998-2003 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
 #include <stdlib.h>
@@ -16,10 +16,41 @@
 #include <windows.h>
 #include <commctrl.h>
 #include <richedit.h>
+#include <windowsx.h>
 
 #include "Platform.h"
 #include "PlatformRes.h"
 #include "UniConversion.h"
+#include "XPM.h"
+
+#ifndef IDC_HAND
+#define IDC_HAND MAKEINTRESOURCE(32649)
+#endif
+
+// Take care of 32/64 bit pointers
+#ifdef GetWindowLongPtr
+static void *PointerFromWindow(HWND hWnd) {
+	return reinterpret_cast<void *>(::GetWindowLongPtr(hWnd, 0));
+}
+static void SetWindowPointer(HWND hWnd, void *ptr) {
+	::SetWindowLongPtr(hWnd, 0, reinterpret_cast<LONG_PTR>(ptr));
+}
+#else
+static void *PointerFromWindow(HWND hWnd) {
+	return reinterpret_cast<void *>(::GetWindowLong(hWnd, 0));
+}
+static void SetWindowPointer(HWND hWnd, void *ptr) {
+	::SetWindowLong(hWnd, 0, reinterpret_cast<LONG>(ptr));
+}
+#endif
+
+static CRITICAL_SECTION crPlatformLock;
+static HINSTANCE hinstPlatformRes = 0;
+static bool onNT = false;
+
+bool IsNT() {
+	return onNT;
+}
 
 Point Point::FromLong(long lpoint) {
 	return Point(static_cast<short>(LOWORD(lpoint)), static_cast<short>(HIWORD(lpoint)));
@@ -103,14 +134,14 @@ void Palette::Allocate(Window &) {
 	}
 }
 
-void SetLogFont(LOGFONT &lf, const char *faceName, int characterSet, int size, bool bold, bool italic) {
+static void SetLogFont(LOGFONT &lf, const char *faceName, int characterSet, int size, bool bold, bool italic) {
 	memset(&lf, 0, sizeof(lf));
 	// The negative is to allow for leading
 	lf.lfHeight = -(abs(size));
 	lf.lfWeight = bold ? FW_BOLD : FW_NORMAL;
 	lf.lfItalic = static_cast<BYTE>(italic ? 1 : 0);
 	lf.lfCharSet = static_cast<BYTE>(characterSet);
-	strcpy(lf.lfFaceName, faceName);
+	strncpy(lf.lfFaceName, faceName, sizeof(lf.lfFaceName));
 }
 
 /**
@@ -118,7 +149,7 @@ void SetLogFont(LOGFONT &lf, const char *faceName, int characterSet, int size, b
  * If one font is the same as another, its hash will be the same, but if the hash is the
  * same then they may still be different.
  */
-int HashFont(const char *faceName, int characterSet, int size, bool bold, bool italic) {
+static int HashFont(const char *faceName, int characterSet, int size, bool bold, bool italic) {
 	return
 		size ^
 		(characterSet << 10) ^
@@ -169,25 +200,30 @@ void FontCached::Release() {
 }
 
 FontID FontCached::FindOrCreate(const char *faceName_, int characterSet_, int size_, bool bold_, bool italic_) {
+	FontID ret = 0;
+	::EnterCriticalSection(&crPlatformLock);
 	int hashFind = HashFont(faceName_, characterSet_, size_, bold_, italic_);
 	for (FontCached *cur=first; cur; cur=cur->next) {
 		if ((cur->hash == hashFind) &&
 			cur->SameAs(faceName_, characterSet_, size_, bold_, italic_)) {
 			cur->usage++;
-			return cur->id;
+			ret = cur->id;
 		}
 	}
-	FontCached *fc = new FontCached(faceName_, characterSet_, size_, bold_, italic_);
-	if (fc) {
-		fc->next = first;
-		first = fc;
-		return fc->id;
-	} else {
-		return 0;
+	if (ret == 0) {
+		FontCached *fc = new FontCached(faceName_, characterSet_, size_, bold_, italic_);
+		if (fc) {
+			fc->next = first;
+			first = fc;
+			ret = fc->id;
+		}
 	}
+	::LeaveCriticalSection(&crPlatformLock);
+	return ret;
 }
 
 void FontCached::ReleaseId(FontID id_) {
+	::EnterCriticalSection(&crPlatformLock);
 	FontCached **pcur=&first;
 	for (FontCached *cur=first; cur; cur=cur->next) {
 		if (cur->id == id_) {
@@ -198,10 +234,11 @@ void FontCached::ReleaseId(FontID id_) {
 				cur->next = 0;
 				delete cur;
 			}
-			return;
+			break;
 		}
 		pcur=&cur->next;
 	}
+	::LeaveCriticalSection(&crPlatformLock);
 }
 
 Font::Font() {
@@ -213,7 +250,8 @@ Font::~Font() {
 
 #define FONTS_CACHED
 
-void Font::Create(const char *faceName, int characterSet, int size, bool bold, bool italic) {
+void Font::Create(const char *faceName, int characterSet, int size,
+	bool bold, bool italic, bool) {
 	Release();
 #ifndef FONTS_CACHED
 	LOGFONT lf;
@@ -248,6 +286,9 @@ class SurfaceImpl : public Surface {
 	HBITMAP bitmap;
 	HBITMAP bitmapOld;
 	HPALETTE paletteOld;
+	int maxWidthMeasure;
+	int maxLenText;
+
 	void BrushColor(ColourAllocated back);
 	void SetFont(Font &font_);
 
@@ -258,9 +299,9 @@ public:
 	SurfaceImpl();
 	virtual ~SurfaceImpl();
 
-	void Init();
-	void Init(SurfaceID sid);
-	void InitPixMap(int width, int height, Surface *surface_);
+	void Init(WindowID wid);
+	void Init(SurfaceID sid, WindowID wid);
+	void InitPixMap(int width, int height, Surface *surface_, WindowID wid);
 
 	void Release();
 	bool Initialised();
@@ -279,6 +320,7 @@ public:
 
 	void DrawTextNoClip(PRectangle rc, Font &font_, int ybase, const char *s, int len, ColourAllocated fore, ColourAllocated back);
 	void DrawTextClipped(PRectangle rc, Font &font_, int ybase, const char *s, int len, ColourAllocated fore, ColourAllocated back);
+	void DrawTextTransparent(PRectangle rc, Font &font_, int ybase, const char *s, int len, ColourAllocated fore);
 	void MeasureWidths(Font &font_, const char *s, int len, int *positions);
 	int WidthText(Font &font_, const char *s, int len);
 	int WidthChar(Font &font_, char ch);
@@ -294,6 +336,7 @@ public:
 	void FlushCachedState();
 
 	void SetUnicodeMode(bool unicodeMode_);
+	void SetDBCSMode(int codePage);
 };
 
 SurfaceImpl::SurfaceImpl() :
@@ -304,6 +347,11 @@ SurfaceImpl::SurfaceImpl() :
 	font(0), 	fontOld(0),
 	bitmap(0), bitmapOld(0),
 	paletteOld(0) {
+	// Windows 9x has only a 16 bit coordinate system so break after 30000 pixels
+	maxWidthMeasure = IsNT() ? 1000000 : 30000;
+	// There appears to be a 16 bit string length limit in GDI on NT and a limit of
+	// 8192 characters on Windows 95.
+	maxLenText = IsNT() ? 65535 : 8192;
 }
 
 SurfaceImpl::~SurfaceImpl() {
@@ -337,7 +385,7 @@ void SurfaceImpl::Release() {
 	bitmap = 0;
 	if (paletteOld) {
 		// Fonts are not deleted as they are owned by a Palette object
-		::SelectPalette(reinterpret_cast<HDC>(hdc), 
+		::SelectPalette(reinterpret_cast<HDC>(hdc),
 			reinterpret_cast<HPALETTE>(paletteOld), TRUE);
 		paletteOld = 0;
 	}
@@ -352,20 +400,20 @@ bool SurfaceImpl::Initialised() {
 	return hdc != 0;
 }
 
-void SurfaceImpl::Init() {
+void SurfaceImpl::Init(WindowID) {
 	Release();
 	hdc = ::CreateCompatibleDC(NULL);
 	hdcOwned = true;
 	::SetTextAlign(reinterpret_cast<HDC>(hdc), TA_BASELINE);
 }
 
-void SurfaceImpl::Init(SurfaceID sid) {
+void SurfaceImpl::Init(SurfaceID sid, WindowID) {
 	Release();
 	hdc = reinterpret_cast<HDC>(sid);
 	::SetTextAlign(reinterpret_cast<HDC>(hdc), TA_BASELINE);
 }
 
-void SurfaceImpl::InitPixMap(int width, int height, Surface *surface_) {
+void SurfaceImpl::InitPixMap(int width, int height, Surface *surface_, WindowID) {
 	Release();
 	hdc = ::CreateCompatibleDC(static_cast<SurfaceImpl *>(surface_)->hdc);
 	hdcOwned = true;
@@ -472,14 +520,14 @@ void SurfaceImpl::Ellipse(PRectangle rc, ColourAllocated fore, ColourAllocated b
 }
 
 void SurfaceImpl::Copy(PRectangle rc, Point from, Surface &surfaceSource) {
-	::BitBlt(hdc, 
+	::BitBlt(hdc,
 		rc.left, rc.top, rc.Width(), rc.Height(),
 		static_cast<SurfaceImpl &>(surfaceSource).hdc, from.x, from.y, SRCCOPY);
 }
 
-#define MAX_US_LEN 5000
+#define MAX_US_LEN 10000
 
-void SurfaceImpl::DrawTextNoClip(PRectangle rc, Font &font_, int ybase, const char *s, int len, 
+void SurfaceImpl::DrawTextNoClip(PRectangle rc, Font &font_, int ybase, const char *s, int len,
 	ColourAllocated fore, ColourAllocated back) {
 	SetFont(font_);
 	::SetTextColor(hdc, fore.AsLong());
@@ -487,15 +535,16 @@ void SurfaceImpl::DrawTextNoClip(PRectangle rc, Font &font_, int ybase, const ch
 	RECT rcw = RectFromPRectangle(rc);
 	if (unicodeMode) {
 		wchar_t tbuf[MAX_US_LEN];
-		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t));
+		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t)-1);
 		tbuf[tlen] = L'\0';
 		::ExtTextOutW(hdc, rc.left, ybase, ETO_OPAQUE, &rcw, tbuf, tlen, NULL);
 	} else {
-		::ExtTextOut(hdc, rc.left, ybase, ETO_OPAQUE, &rcw, s, len, NULL);
+		::ExtTextOut(hdc, rc.left, ybase, ETO_OPAQUE, &rcw, s,
+			Platform::Minimum(len, maxLenText), NULL);
 	}
 }
 
-void SurfaceImpl::DrawTextClipped(PRectangle rc, Font &font_, int ybase, const char *s, int len, 
+void SurfaceImpl::DrawTextClipped(PRectangle rc, Font &font_, int ybase, const char *s, int len,
 	ColourAllocated fore, ColourAllocated back) {
 	SetFont(font_);
 	::SetTextColor(hdc, fore.AsLong());
@@ -503,11 +552,36 @@ void SurfaceImpl::DrawTextClipped(PRectangle rc, Font &font_, int ybase, const c
 	RECT rcw = RectFromPRectangle(rc);
 	if (unicodeMode) {
 		wchar_t tbuf[MAX_US_LEN];
-		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t));
+		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t)-1);
 		tbuf[tlen] = L'\0';
 		::ExtTextOutW(hdc, rc.left, ybase, ETO_OPAQUE | ETO_CLIPPED, &rcw, tbuf, tlen, NULL);
 	} else {
-		::ExtTextOut(hdc, rc.left, ybase, ETO_OPAQUE | ETO_CLIPPED, &rcw, s, len, NULL);
+		::ExtTextOut(hdc, rc.left, ybase, ETO_OPAQUE | ETO_CLIPPED, &rcw, s,
+			Platform::Minimum(len, maxLenText), NULL);
+	}
+}
+
+void SurfaceImpl::DrawTextTransparent(PRectangle rc, Font &font_, int ybase, const char *s, int len,
+	ColourAllocated fore) {
+	// Avoid drawing spaces in transparent mode
+	for (int i=0;i<len;i++) {
+		if (s[i] != ' ') {
+			SetFont(font_);
+			::SetTextColor(hdc, fore.AsLong());
+			::SetBkMode(hdc, TRANSPARENT);
+			RECT rcw = RectFromPRectangle(rc);
+			if (unicodeMode) {
+				wchar_t tbuf[MAX_US_LEN];
+				int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t)-1);
+				tbuf[tlen] = L'\0';
+				::ExtTextOutW(hdc, rc.left, ybase, 0, &rcw, tbuf, tlen, NULL);
+			} else {
+				::ExtTextOut(hdc, rc.left, ybase, 0, &rcw, s,
+					Platform::Minimum(len,maxLenText), NULL);
+			}
+			::SetBkMode(hdc, OPAQUE);
+			return;
+		}
 	}
 }
 
@@ -516,11 +590,11 @@ int SurfaceImpl::WidthText(Font &font_, const char *s, int len) {
 	SIZE sz={0,0};
 	if (unicodeMode) {
 		wchar_t tbuf[MAX_US_LEN];
-		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t));
+		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t)-1);
 		tbuf[tlen] = L'\0';
 		::GetTextExtentPoint32W(hdc, tbuf, tlen, &sz);
 	} else {
-		::GetTextExtentPoint32(hdc, s, len, &sz);
+		::GetTextExtentPoint32(hdc, s, Platform::Minimum(len, maxLenText), &sz);
 	}
 	return sz.cx;
 }
@@ -531,11 +605,11 @@ void SurfaceImpl::MeasureWidths(Font &font_, const char *s, int len, int *positi
 	int fit = 0;
 	if (unicodeMode) {
 		wchar_t tbuf[MAX_US_LEN];
-		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t));
+		int tlen = UCS2FromUTF8(s, len, tbuf, sizeof(tbuf)/sizeof(wchar_t)-1);
 		tbuf[tlen] = L'\0';
 		int poses[MAX_US_LEN];
 		fit = tlen;
-		if (!::GetTextExtentExPointW(hdc, tbuf, tlen, 30000, &fit, poses, &sz)) {
+		if (!::GetTextExtentExPointW(hdc, tbuf, tlen, maxWidthMeasure, &fit, poses, &sz)) {
 			// Likely to have failed because on Windows 9x where function not available
 			// So measure the character widths by measuring each initial substring
 			// Turns a linear operation into a qudratic but seems fast enough on test files
@@ -548,7 +622,7 @@ void SurfaceImpl::MeasureWidths(Font &font_, const char *s, int len, int *positi
 		int ui=0;
 		const unsigned char *us = reinterpret_cast<const unsigned char *>(s);
 		int i=0;
-		while (i<len) {
+		while (ui<fit) {
 			unsigned char uch = us[i];
 			positions[i++] = poses[ui];
 			if (uch >= 0x80) {
@@ -561,9 +635,15 @@ void SurfaceImpl::MeasureWidths(Font &font_, const char *s, int len, int *positi
 			}
 			ui++;
 		}
-		positions[i] = sz.cx;
+		int lastPos = 0;
+		if (i > 0)
+			lastPos = positions[i-1];
+		while (i<len) {
+			positions[i++] = lastPos;
+		}
 	} else {
-		if (!::GetTextExtentExPoint(hdc, s, len, 30000, &fit, positions, &sz)) {
+		if (!::GetTextExtentExPoint(hdc, s, Platform::Minimum(len, maxLenText),
+			maxWidthMeasure, &fit, positions, &sz)) {
 			// Eeek - a NULL DC or other foolishness could cause this.
 			// The least we can do is set the positions to zero!
 			memset(positions, 0, len * sizeof(*positions));
@@ -632,7 +712,7 @@ int SurfaceImpl::SetPalette(Palette *pal, bool inBackGround) {
 	paletteOld = 0;
 	int changes = 0;
 	if (pal->allowRealization) {
-		paletteOld = ::SelectPalette(hdc, 
+		paletteOld = ::SelectPalette(hdc,
 			reinterpret_cast<HPALETTE>(pal->hpal), inBackGround);
 		changes = ::RealizePalette(hdc);
 	}
@@ -653,6 +733,10 @@ void SurfaceImpl::SetUnicodeMode(bool unicodeMode_) {
 	unicodeMode=unicodeMode_;
 }
 
+void SurfaceImpl::SetDBCSMode(int) {
+	// No action on window as automatically handled by system.
+}
+
 Surface *Surface::Allocate() {
 	return new SurfaceImpl;
 }
@@ -677,23 +761,36 @@ PRectangle Window::GetPosition() {
 }
 
 void Window::SetPosition(PRectangle rc) {
-	::SetWindowPos(reinterpret_cast<HWND>(id), 
-		0, rc.left, rc.top, rc.Width(), rc.Height(), 0);
+	::SetWindowPos(reinterpret_cast<HWND>(id),
+		0, rc.left, rc.top, rc.Width(), rc.Height(), SWP_NOZORDER|SWP_NOACTIVATE);
 }
 
-void Window::SetPositionRelative(PRectangle rc, Window) {
+void Window::SetPositionRelative(PRectangle rc, Window w) {
+	LONG style = ::GetWindowLong(reinterpret_cast<HWND>(id), GWL_STYLE);
+	if (style & WS_POPUP) {
+		RECT rcOther;
+		::GetWindowRect(reinterpret_cast<HWND>(w.GetID()), &rcOther);
+		rc.Move(rcOther.left, rcOther.top);
+		if (rc.left < 0) {
+			rc.Move(-rc.left,0);
+		}
+		if (rc.top < 0) {
+			rc.Move(0,-rc.top);
+		}
+	}
 	SetPosition(rc);
 }
 
 PRectangle Window::GetClientPosition() {
-	RECT rc;
-	::GetClientRect(reinterpret_cast<HWND>(id), &rc);
+	RECT rc={0,0,0,0};
+	if (id)
+		::GetClientRect(reinterpret_cast<HWND>(id), &rc);
 	return  PRectangle(rc.left, rc.top, rc.right, rc.bottom);
 }
 
 void Window::Show(bool show) {
 	if (show)
-		::ShowWindow(reinterpret_cast<HWND>(id), SW_SHOWNORMAL);
+		::ShowWindow(reinterpret_cast<HWND>(id), SW_SHOWNOACTIVATE);
 	else
 		::ShowWindow(reinterpret_cast<HWND>(id), SW_HIDE);
 }
@@ -716,8 +813,6 @@ void Window::SetFont(Font &font) {
 		reinterpret_cast<WPARAM>(font.GetID()), 0);
 }
 
-HINSTANCE hinstPlatformRes = 0;
-
 void Window::SetCursor(Cursor curs) {
 	switch (curs) {
 	case cursorText:
@@ -735,6 +830,9 @@ void Window::SetCursor(Cursor curs) {
 	case cursorVert:
 		::SetCursor(::LoadCursor(NULL,IDC_SIZENS));
 		break;
+	case cursorHand:
+		::SetCursor(::LoadCursor(NULL,IDC_HAND));
+		break;
 	case cursorReverseArrow: {
 			if (!hinstPlatformRes)
 				hinstPlatformRes = ::GetModuleHandle("Scintilla");
@@ -742,7 +840,11 @@ void Window::SetCursor(Cursor curs) {
 				hinstPlatformRes = ::GetModuleHandle("SciLexer");
 			if (!hinstPlatformRes)
 				hinstPlatformRes = ::GetModuleHandle(NULL);
-			::SetCursor(::LoadCursor(hinstPlatformRes, MAKEINTRESOURCE(IDC_MARGIN)));
+			HCURSOR hcursor = ::LoadCursor(hinstPlatformRes, MAKEINTRESOURCE(IDC_MARGIN));
+			if (hcursor)
+				::SetCursor(hcursor);
+			else
+				::SetCursor(::LoadCursor(NULL,IDC_ARROW));
 		}
 		break;
 	case cursorArrow:
@@ -756,39 +858,146 @@ void Window::SetTitle(const char *s) {
 	::SetWindowText(reinterpret_cast<HWND>(id), s);
 }
 
-ListBox::ListBox() : desiredVisibleRows(5), maxItemCharacters(0), aveCharWidth(8) {
+class LineToType {
+	int *data;
+	int len;
+	int maximum;
+public:
+	LineToType() :data(0), len(0), maximum(0) {
+	}
+	~LineToType() {
+		Clear();
+	}
+	void Clear() {
+		delete []data;
+		data = 0;
+		len = 0;
+		maximum = 0;
+	}
+	void Add(int index, int value) {
+		if (index >= maximum) {
+			if (index >= len) {
+				int lenNew = (index+1) * 2;
+				int *dataNew = new int[lenNew];
+				for (int i=0; i<maximum; i++) {
+					dataNew[i] = data[i];
+				}
+				delete []data;
+				data = dataNew;
+				len = lenNew;
+			}
+			while (maximum < index) {
+				data[maximum] = 0;
+				maximum++;
+			}
+		}
+		data[index] = value;
+		if (index == maximum) {
+			maximum++;
+		}
+	}
+	int Get(int index) {
+		if (index < maximum) {
+			return data[index];
+		} else {
+			return 0;
+		}
+	}
+};
+
+const char ListBoxX_ClassName[] = "ListBoxX";
+
+ListBox::ListBox() {
 }
 
 ListBox::~ListBox() {
 }
 
-void ListBox::Create(Window &parent, int ctrlID) {
-	HINSTANCE hinstanceParent = reinterpret_cast<HINSTANCE>(
-		::GetWindowLong(reinterpret_cast<HWND>(parent.GetID()),GWL_HINSTANCE));
+class ListBoxX : public ListBox {
+	int lineHeight;
+	FontID fontCopy;
+	XPMSet xset;
+	LineToType ltt;
+	HWND lb;
+	bool unicodeMode;
+	int desiredVisibleRows;
+	unsigned int maxItemCharacters;
+	unsigned int aveCharWidth;
+public:
+	ListBoxX() : lineHeight(10), fontCopy(0), lb(0), unicodeMode(false),
+		desiredVisibleRows(5), maxItemCharacters(0), aveCharWidth(8){
+	}
+	virtual ~ListBoxX() {
+		if (fontCopy) {
+			::DeleteObject(fontCopy);
+			fontCopy = 0;
+		}
+	}
+	virtual void SetFont(Font &font);
+	virtual void Create(Window &parent, int ctrlID, int lineHeight_, bool unicodeMode_);
+	virtual void SetAverageCharWidth(int width);
+	virtual void SetVisibleRows(int rows);
+	virtual PRectangle GetDesiredRect();
+	int IconWidth();
+	virtual int CaretFromEdge();
+	virtual void Clear();
+	virtual void Append(char *s, int type = -1);
+	virtual int Length();
+	virtual void Select(int n);
+	virtual int GetSelection();
+	virtual int Find(const char *prefix);
+	virtual void GetValue(int n, char *value, int len);
+	virtual void RegisterImage(int type, const char *xpm_data);
+	virtual void ClearRegisteredImages();
+	virtual void SetDoubleClickAction(CallBackAction, void *) {
+	}
+	void Draw(DRAWITEMSTRUCT *pDrawItem);
+	long WndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam);
+	static long PASCAL StaticWndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam);
+};
+
+ListBox *ListBox::Allocate() {
+	ListBoxX *lb = new ListBoxX();
+	return lb;
+}
+
+void ListBoxX::Create(Window &parent, int ctrlID, int lineHeight_, bool unicodeMode_) {
+	lineHeight = lineHeight_;
+	unicodeMode = unicodeMode_;
+	HINSTANCE hinstanceParent = GetWindowInstance(reinterpret_cast<HWND>(parent.GetID()));
 	id = ::CreateWindowEx(
-		WS_EX_WINDOWEDGE, "listbox", "",
-		WS_CHILD | WS_THICKFRAME | WS_VSCROLL | LBS_NOTIFY,
-		100,100, 150,80, reinterpret_cast<HWND>(parent.GetID()), 
+		WS_EX_WINDOWEDGE, ListBoxX_ClassName, "",
+		WS_CHILD | WS_THICKFRAME,
+		100,100, 150,80, reinterpret_cast<HWND>(parent.GetID()),
 		reinterpret_cast<HMENU>(ctrlID),
-		hinstanceParent, 
-		0);
+		hinstanceParent,
+		this);
+	lb = ::GetWindow(reinterpret_cast<HWND>(id), GW_CHILD);
 }
 
-void ListBox::SetFont(Font &font) {
-	Window::SetFont(font);
+void ListBoxX::SetFont(Font &font) {
+	LOGFONT lf;
+	if (0 != ::GetObject(font.GetID(), sizeof(lf), &lf)) {
+		if (fontCopy) {
+			::DeleteObject(fontCopy);
+			fontCopy = 0;
+		}
+		fontCopy = ::CreateFontIndirect(&lf);
+		::SendMessage(lb, WM_SETFONT, reinterpret_cast<WPARAM>(fontCopy), 0);
+	}
 }
 
-void ListBox::SetAverageCharWidth(int width) {
+void ListBoxX::SetAverageCharWidth(int width) {
 	aveCharWidth = width;
 }
 
-void ListBox::SetVisibleRows(int rows) {
+void ListBoxX::SetVisibleRows(int rows) {
 	desiredVisibleRows = rows;
 }
 
-PRectangle ListBox::GetDesiredRect() {
+PRectangle ListBoxX::GetDesiredRect() {
 	PRectangle rcDesired = GetPosition();
-	int itemHeight = Window_SendMessage(this, LB_GETITEMHEIGHT, 0);
+	int itemHeight = ::SendMessage(lb, LB_GETITEMHEIGHT, 0, 0);
 	int rows = Length();
 	if ((rows == 0) || (rows > desiredVisibleRows))
 		rows = desiredVisibleRows;
@@ -799,45 +1008,57 @@ PRectangle ListBox::GetDesiredRect() {
 		width = 12;
 	rcDesired.right = rcDesired.left + width * (aveCharWidth+aveCharWidth/3);
 	if (Length() > rows)
-		rcDesired.right = rcDesired.right + ::GetSystemMetrics(SM_CXVSCROLL);
+		rcDesired.right += ::GetSystemMetrics(SM_CXVSCROLL);
+	rcDesired.right += IconWidth();
 	return rcDesired;
 }
 
-void ListBox::Clear() {
-	Window_SendMessage(this, LB_RESETCONTENT);
-	maxItemCharacters = 0;
+int ListBoxX::IconWidth() {
+	return xset.GetWidth() + 2;
 }
 
-void ListBox::Append(char *s) {
-	Window_SendMessage(this, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(s));
-	size_t len = strlen(s);
+int ListBoxX::CaretFromEdge() {
+	return 4 + IconWidth();
+}
+
+void ListBoxX::Clear() {
+	::SendMessage(lb, LB_RESETCONTENT, 0, 0);
+	maxItemCharacters = 0;
+	ltt.Clear();
+}
+
+void ListBoxX::Append(char *s, int type) {
+	::SendMessage(lb, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(s));
+	unsigned int len = static_cast<unsigned int>(strlen(s));
 	if (maxItemCharacters < len)
 		maxItemCharacters = len;
+	int count = ::SendMessage(lb, LB_GETCOUNT, 0, 0);
+	ltt.Add(count-1, type);
 }
 
-int ListBox::Length() {
-	return Window_SendMessage(this, LB_GETCOUNT);
+int ListBoxX::Length() {
+	return ::SendMessage(lb, LB_GETCOUNT, 0, 0);
 }
 
-void ListBox::Select(int n) {
-	Window_SendMessage(this, LB_SETCURSEL, n);
+void ListBoxX::Select(int n) {
+	::SendMessage(lb, LB_SETCURSEL, n, 0);
 }
 
-int ListBox::GetSelection() {
-	return Window_SendMessage(this, LB_GETCURSEL);
+int ListBoxX::GetSelection() {
+	return ::SendMessage(lb, LB_GETCURSEL, 0, 0);
 }
 
-int ListBox::Find(const char *prefix) {
-	return Window_SendMessage(this, LB_FINDSTRING, static_cast<WPARAM>(-1),
+int ListBoxX::Find(const char *prefix) {
+	return ::SendMessage(lb, LB_FINDSTRING, static_cast<WPARAM>(-1),
 		reinterpret_cast<LPARAM>(prefix));
 }
 
-void ListBox::GetValue(int n, char *value, int len) {
-	int lenText = Window_SendMessage(this, LB_GETTEXTLEN, n);
+void ListBoxX::GetValue(int n, char *value, int len) {
+	int lenText = ::SendMessage(lb, LB_GETTEXTLEN, n, 0);
 	if ((len > 0) && (lenText > 0)){
 		char *text = new char[len+1];
 		if (text) {
-			Window_SendMessage(this, LB_GETTEXT, n, reinterpret_cast<LPARAM>(text));
+			::SendMessage(lb, LB_GETTEXT, n, reinterpret_cast<LPARAM>(text));
 			strncpy(value, text, len);
 			value[len-1] = '\0';
 			delete []text;
@@ -849,8 +1070,154 @@ void ListBox::GetValue(int n, char *value, int len) {
 	}
 }
 
-void ListBox::Sort() {
-	// Windows keeps sorted so no need to sort
+void ListBoxX::RegisterImage(int type, const char *xpm_data) {
+	xset.Add(type, xpm_data);
+}
+
+void ListBoxX::ClearRegisteredImages() {
+	xset.Clear();
+}
+
+void ListBoxX::Draw(DRAWITEMSTRUCT *pDrawItem) {
+	if ((pDrawItem->itemAction == ODA_SELECT) || (pDrawItem->itemAction == ODA_DRAWENTIRE)) {
+		char text[1000];
+		int len = ::SendMessage(pDrawItem->hwndItem, LB_GETTEXTLEN, pDrawItem->itemID, 0);
+		if (len < static_cast<int>(sizeof(text))) {
+			::SendMessage(pDrawItem->hwndItem, LB_GETTEXT, pDrawItem->itemID, reinterpret_cast<LPARAM>(text));
+		} else {
+			len = 0;
+			text[0] = '\0';
+		}
+		int pixId = ltt.Get(pDrawItem->itemID);
+		XPM *pxpm = xset.Get(pixId);
+		if (pDrawItem->itemState & ODS_SELECTED) {
+			::SetBkColor(pDrawItem->hDC, ::GetSysColor(COLOR_HIGHLIGHT));
+			::SetTextColor(pDrawItem->hDC, ::GetSysColor(COLOR_HIGHLIGHTTEXT));
+		} else {
+			::SetBkColor(pDrawItem->hDC, ::GetSysColor(COLOR_WINDOW));
+			::SetTextColor(pDrawItem->hDC, ::GetSysColor(COLOR_WINDOWTEXT));
+		}
+		int widthPix = xset.GetWidth() + 2;
+		if (unicodeMode) {
+			wchar_t tbuf[MAX_US_LEN];
+			int tlen = UCS2FromUTF8(text, len, tbuf, sizeof(tbuf)/sizeof(wchar_t)-1);
+			tbuf[tlen] = L'\0';
+			::ExtTextOutW(pDrawItem->hDC, pDrawItem->rcItem.left+widthPix+1, pDrawItem->rcItem.top,
+				ETO_OPAQUE, &(pDrawItem->rcItem), tbuf, tlen, NULL);
+		} else {
+			::ExtTextOut(pDrawItem->hDC, pDrawItem->rcItem.left+widthPix+1, pDrawItem->rcItem.top,
+				ETO_OPAQUE, &(pDrawItem->rcItem), text,
+				len, NULL);
+		}
+		if (pxpm) {
+			Surface *surfaceItem = Surface::Allocate();
+			if (surfaceItem) {
+				surfaceItem->Init(pDrawItem->hDC, pDrawItem->hwndItem);
+				//surfaceItem->SetUnicodeMode(unicodeMode);
+				//surfaceItem->SetDBCSMode(codePage);
+				int left = pDrawItem->rcItem.left;
+				PRectangle rc(left + 1, pDrawItem->rcItem.top,
+					left + 1 + widthPix, pDrawItem->rcItem.bottom);
+				pxpm->Draw(surfaceItem, rc);
+				delete surfaceItem;
+				::SetTextAlign(pDrawItem->hDC, TA_TOP);
+			}
+		}
+        if (pDrawItem->itemState & ODS_SELECTED) {
+			::DrawFocusRect(pDrawItem->hDC, &(pDrawItem->rcItem));
+        }
+	}
+}
+
+long ListBoxX::WndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam) {
+	switch (iMessage) {
+	case WM_CREATE: {
+		HWND parent = ::GetParent(hWnd);
+		HINSTANCE hinstanceParent = GetWindowInstance(parent);
+		CREATESTRUCT *pCreate = reinterpret_cast<CREATESTRUCT *>(lParam);
+		::CreateWindowEx(
+			WS_EX_WINDOWEDGE, "listbox", "",
+			WS_CHILD | WS_VSCROLL | WS_VISIBLE |
+			LBS_OWNERDRAWFIXED | LBS_HASSTRINGS | LBS_NOTIFY,
+			0, 0, 150,80, hWnd,
+			pCreate->hMenu,
+			hinstanceParent,
+			0);
+		}
+		break;
+	case WM_SIZE: {
+			HWND lb = ::GetWindow(hWnd, GW_CHILD);
+			::SetWindowPos(lb, 0, 0,0, LOWORD(lParam), HIWORD(lParam), 0);
+		}
+		break;
+	case WM_PAINT: {
+			PAINTSTRUCT ps;
+			::BeginPaint(hWnd, &ps);
+			::EndPaint(hWnd, &ps);
+		}
+		break;
+	case WM_COMMAND: {
+			HWND parent = ::GetParent(hWnd);
+			::SendMessage(parent, iMessage, wParam, lParam);
+		}
+		break;
+
+	case WM_MEASUREITEM: {
+			MEASUREITEMSTRUCT *pMeasureItem = reinterpret_cast<MEASUREITEMSTRUCT *>(lParam);
+			pMeasureItem->itemHeight = lineHeight;
+			if (pMeasureItem->itemHeight < static_cast<unsigned int>(xset.GetHeight())) {
+				pMeasureItem->itemHeight = xset.GetHeight();
+			}
+		}
+		break;
+	case WM_DRAWITEM:
+		Draw(reinterpret_cast<DRAWITEMSTRUCT *>(lParam));
+		break;
+	case WM_DESTROY:
+		::SetWindowLong(hWnd, 0, 0);
+		return ::DefWindowProc(hWnd, iMessage, wParam, lParam);
+	default:
+		return ::DefWindowProc(hWnd, iMessage, wParam, lParam);
+	}
+
+	return 0;
+}
+
+long PASCAL ListBoxX::StaticWndProc(
+    HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam) {
+	if (iMessage == WM_CREATE) {
+		CREATESTRUCT *pCreate = reinterpret_cast<CREATESTRUCT *>(lParam);
+		SetWindowPointer(hWnd, pCreate->lpCreateParams);
+	}
+	// Find C++ object associated with window.
+	ListBoxX *lbx = reinterpret_cast<ListBoxX *>(PointerFromWindow(hWnd));
+	if (lbx) {
+		return lbx->WndProc(hWnd, iMessage, wParam, lParam);
+	} else {
+		return ::DefWindowProc(hWnd, iMessage, wParam, lParam);
+	}
+}
+
+static bool ListBoxX_Register() {
+	WNDCLASSEX wndclassc;
+	wndclassc.cbSize = sizeof(wndclassc);
+	wndclassc.style = CS_GLOBALCLASS | CS_HREDRAW | CS_VREDRAW;
+	wndclassc.cbClsExtra = 0;
+	wndclassc.cbWndExtra = sizeof(ListBoxX *);
+	wndclassc.hInstance = hinstPlatformRes;
+	wndclassc.hIcon = NULL;
+	wndclassc.hbrBackground = NULL;
+	wndclassc.lpszMenuName = NULL;
+	wndclassc.lpfnWndProc = ListBoxX::StaticWndProc;
+	wndclassc.hCursor = ::LoadCursor(NULL, IDC_ARROW);
+	wndclassc.lpszClassName = ListBoxX_ClassName;
+	wndclassc.hIconSm = 0;
+
+	return ::RegisterClassEx(&wndclassc) != 0;
+}
+
+bool ListBoxX_Unregister() {
+	return ::UnregisterClass(ListBoxX_ClassName, hinstPlatformRes) != 0;
 }
 
 Menu::Menu() : id(0) {
@@ -868,8 +1235,8 @@ void Menu::Destroy() {
 }
 
 void Menu::Show(Point pt, Window &w) {
-	::TrackPopupMenu(reinterpret_cast<HMENU>(id), 
-		0, pt.x - 4, pt.y, 0, 
+	::TrackPopupMenu(reinterpret_cast<HMENU>(id),
+		0, pt.x - 4, pt.y, 0,
 		reinterpret_cast<HWND>(w.GetID()), NULL);
 	Destroy();
 }
@@ -921,6 +1288,37 @@ double ElapsedTime::Duration(bool reset) {
 	return result;
 }
 
+class DynamicLibraryImpl : public DynamicLibrary {
+protected:
+	HMODULE h;
+public:
+	DynamicLibraryImpl(const char *modulePath) {
+		h = ::LoadLibrary(modulePath);
+	}
+
+	virtual ~DynamicLibraryImpl() {
+		if (h != NULL)
+			::FreeLibrary(h);
+	}
+
+	// Use GetProcAddress to get a pointer to the relevant function.
+	virtual Function FindFunction(const char *name) {
+		if (h != NULL) {
+			return static_cast<Function>(
+				(void *)(::GetProcAddress(h, name)));
+		} else
+			return NULL;
+	}
+
+	virtual bool IsValid() {
+		return h != NULL;
+	}
+};
+
+DynamicLibrary *DynamicLibrary::Load(const char *modulePath) {
+	return static_cast<DynamicLibrary *>( new DynamicLibraryImpl(modulePath) );
+}
+
 ColourDesired Platform::Chrome() {
 	return ::GetSysColor(COLOR_3DFACE);
 }
@@ -941,6 +1339,10 @@ unsigned int Platform::DoubleClickTime() {
 	return ::GetDoubleClickTime();
 }
 
+bool Platform::MouseButtonBounce() {
+	return false;
+}
+
 void Platform::DebugDisplay(const char *s) {
 	::OutputDebugString(s);
 }
@@ -953,10 +1355,23 @@ long Platform::SendScintilla(WindowID w, unsigned int msg, unsigned long wParam,
 	return ::SendMessage(reinterpret_cast<HWND>(w), msg, wParam, lParam);
 }
 
+long Platform::SendScintillaPointer(WindowID w, unsigned int msg, unsigned long wParam, void *lParam) {
+	return ::SendMessage(reinterpret_cast<HWND>(w), msg, wParam,
+		reinterpret_cast<LPARAM>(lParam));
+}
+
 bool Platform::IsDBCSLeadByte(int codePage, char ch) {
 	return ::IsDBCSLeadByteEx(codePage, ch) != 0;
 }
 
+int Platform::DBCSCharLength(int codePage, const char *s) {
+	return (::IsDBCSLeadByteEx(codePage, s[0]) != 0) ? 2 : 1;
+}
+
+int Platform::DBCSCharMaxLength() {
+	return 2;
+}
+
 // These are utility functions not really tied to a platform
 
 int Platform::Minimum(int a, int b) {
@@ -1024,3 +1439,17 @@ int Platform::Clamp(int val, int minVal, int maxVal) {
 		val = minVal;
 	return val;
 }
+
+void Platform_Initialise(void *hInstance) {
+	OSVERSIONINFO osv = {sizeof(OSVERSIONINFO),0,0,0,0,""};
+	::GetVersionEx(&osv);
+	onNT = osv.dwPlatformId == VER_PLATFORM_WIN32_NT;
+	::InitializeCriticalSection(&crPlatformLock);
+	hinstPlatformRes = reinterpret_cast<HINSTANCE>(hInstance);
+	ListBoxX_Register();
+}
+
+void Platform_Finalise() {
+	ListBoxX_Unregister();
+	::DeleteCriticalSection(&crPlatformLock);
+}
