@@ -44,7 +44,7 @@
 #define IDM_ENCODING_UCOOKIE	154
 
 #define MRU_START	16
-#define IMPORT_START 18
+#define IMPORT_START 19
 #define TOOLS_START 3
 
 // Edit
@@ -68,6 +68,8 @@
 #define IDM_BOOKMARK_TOGGLE	222
 #define IDM_BOOKMARK_PREV	223
 #define IDM_BOOKMARK_CLEARALL	224
+#define IDM_BOOKMARK_NEXT_SELECT	225
+#define IDM_BOOKMARK_PREV_SELECT	226
 #define IDM_MATCHBRACE		230
 #define IDM_SELECTTOBRACE	231
 #define IDM_SHOWCALLTIP		232
@@ -156,6 +158,7 @@
 #define IDM_OPENUSERPROPERTIES	461
 #define IDM_OPENGLOBALPROPERTIES	462
 #define IDM_OPENABBREVPROPERTIES	463
+#define IDM_OPENLUAEXTERNALFILE	464
 
 //#define IDM_SELECTIONMARGIN	490
 //#define IDM_BUFFEREDDRAW	491
@@ -216,6 +219,7 @@
 #define IDCURRLINECHAR		247
 #define IDREPLDONE			248
 
+#define IDDOTDOT			249
 
 #define IDPARAMSTART		300
 
