@@ -60,22 +60,22 @@ class EntryMemory {
 	SString entries[sz];
 public:
 	void Insert(const SString &s) {
-		for (int i = 0;i < sz;i++) {
+		for (int i = 0; i < sz; i++) {
 			if (entries[i] == s) {
-				for (int j = i;j > 0;j--) {
+				for (int j = i; j > 0; j--) {
 					entries[j] = entries[j - 1];
 				}
 				entries[0] = s;
 				return;
 			}
 		}
-		for (int k = sz - 1;k > 0;k--) {
+		for (int k = sz - 1; k > 0; k--) {
 			entries[k] = entries[k - 1];
 		}
 		entries[0] = s;
 	}
 	void AppendIfNotPresent(const SString &s) {
-		for (int i = 0;i < sz;i++) {
+		for (int i = 0; i < sz; i++) {
 			if (entries[i] == s) {
 				return;
 			}
@@ -98,7 +98,7 @@ public:
 	}
 	int Length() const {
 		int len = 0;
-		for (int i = 0;i < sz;i++)
+		for (int i = 0; i < sz; i++)
 			if (entries[i].length())
 				len++;
 		return len;
@@ -397,8 +397,8 @@ public:
 	              sdCaseForce = 0x100, sdVisible = 0x200, sdChangeable = 0x400} specified;
 	StyleDefinition(const char *definition);
 	bool ParseStyleDefinition(const char *definition);
-	long ForeAsLong();
-	long BackAsLong();
+	long ForeAsLong() const;
+	long BackAsLong() const;
 };
 
 struct StyleAndWords {
@@ -523,6 +523,7 @@ protected:
 	bool fullScreen;
 	enum { toolMax = 50 };
 	Extension *extender;
+	bool needReadProperties;
 
 	int heightOutput;
 	int heightOutputStartDrag;
@@ -545,6 +546,9 @@ protected:
 	bool callTipIgnoreCase;
 	bool autoCCausedByOnlyOne;
 	SString calltipWordCharacters;
+	SString calltipParametersStart;
+	SString calltipParametersEnd;
+	SString calltipParametersSeparators;
 	SString calltipEndDefinition;
 	SString autoCompleteStartCharacters;
 	SString autoCompleteFillUpCharacters;
@@ -652,10 +656,12 @@ protected:
 	void SetIndentSettings();
 	void SetEol();
 	void New();
-	void Close(bool updateUI = true, bool loadingSession = false);
+	void Close(bool updateUI = true, bool loadingSession = false, bool makingRoomForNew = false);
 	bool IsAbsolutePath(const char *path);
 	bool Exists(const char *dir, const char *path, char *testPath);
-	void OpenFile(bool suppressMessage);
+	void DiscoverEOLSetting();
+	void DiscoverIndentSetting();
+	void OpenFile(int fileSize, bool suppressMessage);
 	virtual void OpenUriList(const char *) {};
 	virtual void AbsolutePath(char *fullPath, const char *basePath, int size) = 0;
 	virtual void FixFilePath();
@@ -711,6 +717,7 @@ protected:
 	void SetSelection(int anchor, int currentPos);
 	//	void SelectionExtend(char *sel, int len, char *notselchar);
 	void GetCTag(char *sel, int len);
+	virtual SString GetRangeInUIEncoding(Window &win, int selStart, int selEnd);
 	SString RangeExtendAndGrab(Window &wCurrent, int &selStart, int &selEnd,
 		bool (*ischarforsel)(char ch), bool stripEol = true);
 	SString SelectionExtend(bool (*ischarforsel)(char ch), bool stripEol = true);
@@ -720,6 +727,7 @@ protected:
 	SString SelectionFilename();
 	void SelectionIntoProperties();
 	void SelectionIntoFind(bool stripEol = true);
+	virtual SString EncodeString(const SString &s);
 	virtual void Find() = 0;
 	virtual int WindowMessageBox(Window &w, const SString &m, int style)=0;
 	void FindMessageBox(const SString &msg);
@@ -772,7 +780,7 @@ protected:
 	void AutomaticIndentation(char ch);
 	void CharAdded(char ch);
 	void SetTextProperties(PropSetFile &ps);
-	void SetFileProperties(PropSet &ps);
+	virtual void SetFileProperties(PropSet &ps) = 0;
 	virtual void UpdateStatusBar(bool bUpdateSlowData);
 	int GetLineLength(int line);
 	int GetCurrentLineNumber();
@@ -808,7 +816,7 @@ protected:
 	void BookmarkDelete(int lineno = -1);
 	bool BookmarkPresent(int lineno = -1);
 	void BookmarkToggle(int lineno = -1);
-	void BookmarkNext(bool forwardScan = true);
+	void BookmarkNext(bool forwardScan = true, bool select = false);
 	void ToggleOutputVisible();
 	virtual void SizeContentWindows() = 0;
 	virtual void SizeSubWindows() = 0;
@@ -855,7 +863,7 @@ protected:
 		const char *param1=0, const char *param2=0);
 	virtual void ReadLocalisation();
 	virtual void ReadPropertiesInitial();
-	void SetMonoFont();
+	void ReadFontProperties();
 	void SetOverrideLanguage(int cmdID);
 	StyleAndWords GetStyleAndWords(const char *base);
 	SString ExtensionFileName();
@@ -865,7 +873,7 @@ protected:
 	void ReadAPI(const SString &fileNameForExtension);
 	SString FindLanguageProperty(const char *pattern, const char *defaultValue="");
 	virtual void ReadProperties();
-	void SetOneStyle(Window &win, int style, const char *s);
+	void SetOneStyle(Window &win, int style, const StyleDefinition &sd);
 	void SetStyleFor(Window &win, const char *language);
 
 	void CheckReload();
@@ -900,6 +908,7 @@ protected:
 	void Trace(const char *s);
 	char *Property(const char *key);
 	void SetProperty(const char *key, const char *val);
+	void UnsetProperty(const char *key);
 	uptr_t GetInstance();
 	void ShutDown();
 	void Perform(const char *actions);
