@@ -501,16 +501,15 @@ bool StyleDefinition::ParseStyleDefinition(const char *definition) {
 	return true;
 }
 
-long StyleDefinition::ForeAsLong() {
+long StyleDefinition::ForeAsLong() const {
 	return ColourFromString(fore);
 }
 
-long StyleDefinition::BackAsLong() {
+long StyleDefinition::BackAsLong() const {
 	return ColourFromString(back);
 }
 
-void SciTEBase::SetOneStyle(Window &win, int style, const char *s) {
-	StyleDefinition sd(s);
+void SciTEBase::SetOneStyle(Window &win, int style, const StyleDefinition &sd) {
 	if (sd.specified & StyleDefinition::sdItalics)
 		Platform::SendScintilla(win.GetID(), SCI_STYLESETITALIC, style, sd.italics ? 1 : 0);
 	if (sd.specified & StyleDefinition::sdBold)
@@ -666,6 +665,7 @@ static const char *propertiesToForward[] = {
 	"html.tags.case.sensitive",
 	"ps.level",
 	"ps.tokenize",
+	"sql.backslash.escapes",
 
 	"lexer.tex.interface.default",
 	"lexer.tex.comment.process",
@@ -680,6 +680,9 @@ static const char *propertiesToForward[] = {
 };
 
 void SciTEBase::ReadProperties() {
+	if (extender)
+		extender->Clear();
+
 	SString fileNameForExtension = ExtensionFileName();
 
 	SString modulePath = props.GetNewExpand("lexerpath.",
@@ -865,6 +868,10 @@ void SciTEBase::ReadProperties() {
 	calltipWordCharacters = FindLanguageProperty("calltip.*.word.characters",
 		"_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ");
 
+	calltipParametersStart = FindLanguageProperty("calltip.*.parameters.start", "(");
+	calltipParametersEnd = FindLanguageProperty("calltip.*.parameters.end", ")");
+	calltipParametersSeparators = FindLanguageProperty("calltip.*.parameters.separators", ",;");
+
 	calltipEndDefinition = FindLanguageProperty("calltip.*.end.definition");
 
 	sprintf(key, "autocomplete.%s.start.characters", language.c_str());
@@ -893,47 +900,14 @@ void SciTEBase::ReadProperties() {
 	int autoCChooseSingle = props.GetInt("autocomplete.choose.single");
 	SendEditor(SCI_AUTOCSETCHOOSESINGLE, autoCChooseSingle),
 
-	SendEditor(SCI_AUTOCSETCANCELATSTART, 0),
-	SendEditor(SCI_AUTOCSETDROPRESTOFWORD, 0),
-
-	// Set styles
-	// For each window set the global default style, then the language default style, then the other global styles, then the other language styles
-
-	SendEditor(SCI_STYLERESETDEFAULT, 0, 0);
-	SendOutput(SCI_STYLERESETDEFAULT, 0, 0);
-
-	sprintf(key, "style.%s.%0d", "*", STYLE_DEFAULT);
-	sval = props.GetNewExpand(key);
-	SetOneStyle(wEditor, STYLE_DEFAULT, sval.c_str());
-	SetOneStyle(wOutput, STYLE_DEFAULT, sval.c_str());
-
-	sprintf(key, "style.%s.%0d", language.c_str(), STYLE_DEFAULT);
-	sval = props.GetNewExpand(key);
-	SetOneStyle(wEditor, STYLE_DEFAULT, sval.c_str());
-
-	SendEditor(SCI_STYLECLEARALL, 0, 0);
-
-	SetStyleFor(wEditor, "*");
-	SetStyleFor(wEditor, language.c_str());
-
-	SendOutput(SCI_STYLECLEARALL, 0, 0);
-
-	sprintf(key, "style.%s.%0d", "errorlist", STYLE_DEFAULT);
-	sval = props.GetNewExpand(key);
-	SetOneStyle(wOutput, STYLE_DEFAULT, sval.c_str());
-
-	SendOutput(SCI_STYLECLEARALL, 0, 0);
-
-	SetStyleFor(wOutput, "*");
-	SetStyleFor(wOutput, "errorlist");
+	SendEditor(SCI_AUTOCSETCANCELATSTART, 0);
+	SendEditor(SCI_AUTOCSETDROPRESTOFWORD, 0);
 
 	if (firstPropertiesRead) {
 		ReadPropertiesInitial();
 	}
 
-	if (useMonoFont) {
-		SetMonoFont();
-	}
+	ReadFontProperties();
 
 	SendEditor(SCI_SETUSEPALETTE, props.GetInt("use.palette"));
 	SendEditor(SCI_SETPRINTMAGNIFICATION, props.GetInt("print.magnification"));
@@ -1148,8 +1122,6 @@ void SciTEBase::ReadProperties() {
 	SendEditor(SCI_SETENDATLASTLINE, props.GetInt("end.at.last.line", 1));
 
 	if (extender) {
-		extender->Clear();
-
 		char defaultDir[MAX_PATH];
 		GetDefaultDirectory(defaultDir, sizeof(defaultDir));
 		char scriptPath[MAX_PATH];
@@ -1172,8 +1144,59 @@ void SciTEBase::ReadProperties() {
 			}
 		}
 	}
-
 	firstPropertiesRead = false;
+	needReadProperties = false;
+}
+
+void SciTEBase::ReadFontProperties() {
+	char key[200];
+	SString sval;
+
+	// Set styles
+	// For each window set the global default style, then the language default style, then the other global styles, then the other language styles
+
+	SendEditor(SCI_STYLERESETDEFAULT, 0, 0);
+	SendOutput(SCI_STYLERESETDEFAULT, 0, 0);
+
+	sprintf(key, "style.%s.%0d", "*", STYLE_DEFAULT);
+	sval = props.GetNewExpand(key);
+	SetOneStyle(wEditor, STYLE_DEFAULT, sval.c_str());
+	SetOneStyle(wOutput, STYLE_DEFAULT, sval.c_str());
+
+	sprintf(key, "style.%s.%0d", language.c_str(), STYLE_DEFAULT);
+	sval = props.GetNewExpand(key);
+	SetOneStyle(wEditor, STYLE_DEFAULT, sval.c_str());
+
+	SendEditor(SCI_STYLECLEARALL, 0, 0);
+
+	SetStyleFor(wEditor, "*");
+	SetStyleFor(wEditor, language.c_str());
+
+	SendOutput(SCI_STYLECLEARALL, 0, 0);
+
+	sprintf(key, "style.%s.%0d", "errorlist", STYLE_DEFAULT);
+	sval = props.GetNewExpand(key);
+	SetOneStyle(wOutput, STYLE_DEFAULT, sval.c_str());
+
+	SendOutput(SCI_STYLECLEARALL, 0, 0);
+
+	SetStyleFor(wOutput, "*");
+	SetStyleFor(wOutput, "errorlist");
+
+	if (useMonoFont) {
+		sval = props.GetExpanded("font.monospace");
+		StyleDefinition sd(sval.c_str());
+		for (int style = 0; style <= STYLE_MAX; style++) {
+			if (style != STYLE_LINENUMBER) {
+				if (sd.specified & StyleDefinition::sdFont) {
+					SendEditorString(SCI_STYLESETFONT, style, sd.font.c_str());
+				}
+				if (sd.specified & StyleDefinition::sdSize) {
+					SendEditor(SCI_STYLESETSIZE, style, sd.size);
+				}
+			}
+		}
+	}
 }
 
 // Properties that are interactively modifiable are only read from the properties file once.
@@ -1211,6 +1234,7 @@ SString SciTEBase::LocaliseString(const char *s, bool retainIfNotFound) {
 	int ellipseIndicator = translation.remove("...");
 	int accessKeyPresent = translation.remove(menuAccessIndicator);
 	translation.lowercase();
+	translation.substitute("\n", "\\n");
 	translation = propsUI.Get(translation.c_str());
 	if (translation.length()) {
 		if (ellipseIndicator)
@@ -1218,6 +1242,7 @@ SString SciTEBase::LocaliseString(const char *s, bool retainIfNotFound) {
 		if (0 == accessKeyPresent)
 			translation.remove("&");
 		translation.substitute("&", menuAccessIndicator);
+		translation.substitute("\\n", "\n");
 	} else {
 		translation = props.Get("translation.missing");
 	}
@@ -1273,7 +1298,7 @@ void SciTEBase::ReadPropertiesInitial() {
 	SendEditor(SCI_SETWRAPMODE, wrap ? SC_WRAP_WORD : SC_WRAP_NONE);
 	SendOutput(SCI_SETWRAPMODE, wrapOutput ? SC_WRAP_WORD : SC_WRAP_NONE);
 
-	useMonoFont = props.GetInt("use.monospaced");
+	useMonoFont = false;
 
 	SString menuLanguageProp = props.GetNewExpand("menu.language");
 	languageItems = 0;
@@ -1345,7 +1370,7 @@ bool SciTEBase::GetDefaultPropertiesFileName(char *pathDefaultProps,
 
 bool SciTEBase::GetAbbrevPropertiesFileName(char *pathAbbrevProps,
         char *pathDefaultDir, unsigned int lenPath) {
-	if (!GetSciteDefaultHome(pathDefaultDir, lenPath)) {
+	if (!GetSciteUserHome(pathDefaultDir, lenPath)) {
 		return false;
 	}
 	return BuildPath(pathAbbrevProps, pathDefaultDir, propAbbrevFileName, lenPath);
@@ -1380,10 +1405,15 @@ void SciTEBase::OpenProperties(int propsFile) {
 		if (GetAbbrevPropertiesFileName(propfile, propdir, sizeof(propfile))) {
 			Open(propfile, ofQuiet);
 		}
-	} else {	// IDM_OPENGLOBALPROPERTIES
+	} else if (propsFile == IDM_OPENGLOBALPROPERTIES) {
 		if (GetDefaultPropertiesFileName(propfile, propdir, sizeof(propfile))) {
 			Open(propfile, ofQuiet);
 		}
+	} else if (propsFile == IDM_OPENLUAEXTERNALFILE) {
+		SString extlua = props.GetExpanded("ext.lua.startup.script");
+		if (extlua.length()) {
+			Open(extlua.c_str(), ofQuiet);
+		}
 	}
 }
 
@@ -1445,6 +1475,8 @@ int SciTEBase::GetMenuCommandAsInt(SString commandName) {
 		{"IDM_BOOKMARK_TOGGLE",	IDM_BOOKMARK_TOGGLE},
 		{"IDM_BOOKMARK_PREV",	IDM_BOOKMARK_PREV},
 		{"IDM_BOOKMARK_CLEARALL",	IDM_BOOKMARK_CLEARALL},
+		{"IDM_BOOKMARK_NEXT_SELECT",	IDM_BOOKMARK_NEXT_SELECT},
+		{"IDM_BOOKMARK_PREV_SELECT",	IDM_BOOKMARK_PREV_SELECT},
 		{"IDM_MATCHBRACE",	IDM_MATCHBRACE},
 		{"IDM_SELECTTOBRACE",	IDM_SELECTTOBRACE},
 		{"IDM_SHOWCALLTIP",	IDM_SHOWCALLTIP},
@@ -1514,6 +1546,7 @@ int SciTEBase::GetMenuCommandAsInt(SString commandName) {
 		{"IDM_OPENUSERPROPERTIES",	IDM_OPENUSERPROPERTIES},
 		{"IDM_OPENGLOBALPROPERTIES",	IDM_OPENGLOBALPROPERTIES},
 		{"IDM_OPENABBREVPROPERTIES",	IDM_OPENABBREVPROPERTIES},
+		{"IDM_OPENLUAEXTERNALFILE",	IDM_OPENLUAEXTERNALFILE},
 		{"IDM_PREVFILE",	IDM_PREVFILE},
 		{"IDM_NEXTFILE",	IDM_NEXTFILE},
 		{"IDM_CLOSEALL",	IDM_CLOSEALL},
