@@ -75,6 +75,51 @@ static void PlayThisSound(
 	}
 }
 
+// Handle Unicode controls (assume strings to be UTF-8 on Windows NT)
+
+// This is a reasonable buffer size for dialog box text conversions
+#define CTL_TEXT_BUF 512
+
+SString GetDlgItemText2(HWND hDlg, int idItem) {
+	SString result = "";
+	WCHAR wsz[CTL_TEXT_BUF];
+	char	msz[CTL_TEXT_BUF];
+
+	if (::IsWindowUnicode(GetDlgItem(hDlg, idItem))) {
+		if (::GetDlgItemTextW(hDlg, idItem, wsz, CTL_TEXT_BUF)) {
+			if (::WideCharToMultiByte(CP_UTF8, 0, wsz, -1, msz, CTL_TEXT_BUF, NULL, NULL))
+				result = msz;
+		}
+	} else {
+		if (::GetDlgItemTextA(hDlg, idItem, msz, CTL_TEXT_BUF))
+			result = msz;
+	}
+
+	//::MessageBox(GetFocus(),result.c_str(),"GetDlgItemText2():out",0);
+
+	return result;
+}
+
+BOOL SetDlgItemText2(HWND hDlg, int idItem, LPCSTR pmsz) {
+	BOOL bSuccess = FALSE;
+	WCHAR wsz[CTL_TEXT_BUF];
+
+	if (!pmsz || *pmsz == 0)
+		return ::SetDlgItemTextA(hDlg, idItem, "");
+
+	//::MessageBox(GetFocus(),pmsz,"SetDlgItemText2():in",0);
+
+	if (::IsWindowUnicode(::GetDlgItem(hDlg, idItem))) {
+		if (::MultiByteToWideChar(CP_UTF8, 0, pmsz, -1, wsz, CTL_TEXT_BUF)) {
+			bSuccess = ::SetDlgItemTextW(hDlg, idItem, wsz);
+		}
+	} else {
+		bSuccess = ::SetDlgItemTextA(hDlg, idItem, pmsz);
+	}
+
+	return bSuccess;
+}
+
 static SciTEWin *Caller(HWND hDlg, UINT message, LPARAM lParam) {
 	if (message == WM_INITDIALOG) {
 		SetWindowLong(hDlg, DWL_USER, lParam);
@@ -123,14 +168,23 @@ void SciTEWin::WarnUser(int warnID) {
 	PlayThisSound(sound, atoi(soundDuration), hMM);
 }
 
-bool SciTEWin::ModelessHandler(MSG *pmsg) {
-	if (wFindReplace.GetID()) {
-		if (::IsDialogMessage(reinterpret_cast<HWND>(wFindReplace.GetID()), pmsg))
+bool DialogHandled(WindowID id, MSG *pmsg) {
+	if (id) {
+		if (::IsDialogMessage(reinterpret_cast<HWND>(id), pmsg))
 			return true;
 	}
-	if (wFindIncrement.GetID()) {
-		if (::IsDialogMessage(reinterpret_cast<HWND>(wFindIncrement.GetID()), pmsg))
-			return true;
+	return false;
+}
+
+bool SciTEWin::ModelessHandler(MSG *pmsg) {
+	if (DialogHandled(wFindReplace.GetID(), pmsg)) {
+		return true;
+	}
+	if (DialogHandled(wFindIncrement.GetID(), pmsg)) {
+		return true;
+	}
+	if (DialogHandled(wFindInFiles.GetID(), pmsg)) {
+		return true;
 	}
 	if (wParameters.GetID()) {
 		bool menuKey = (pmsg->message == WM_KEYDOWN) &&
@@ -159,7 +213,7 @@ int SciTEWin::DoDialog(HINSTANCE hInst, const char *resName, HWND hWnd, DLGPROC
 	if (result == -1) {
 		SString errorNum(::GetLastError());
 		SString msg = LocaliseMessage("Failed to create dialog box: ^0.", errorNum.c_str());
-		MessageBox(hWnd, msg.c_str(), appName, MB_OK | MB_SETFOREGROUND);
+		::MessageBox(hWnd, msg.c_str(), appName, MB_OK | MB_SETFOREGROUND);
 	}
 
 	return result;
@@ -732,11 +786,25 @@ void SciTEWin::PrintSetup() {
 	hDevNames = pdlg.hDevNames;
 }
 
+extern bool IsWindowsNT();
+
 static void FillComboFromMemory(HWND combo, const ComboMemory &mem, bool useTop = false) {
-	for (int i = 0; i < mem.Length(); i++) {
-		//Platform::DebugPrintf("Combo[%0d] = %s\n", i, mem.At(i).c_str());
-		::SendMessage(combo, CB_ADDSTRING, 0,
-		              reinterpret_cast<LPARAM>(mem.At(i).c_str()));
+	::SendMessage(combo, CB_RESETCONTENT, 0, 0);
+	if (::IsWindowUnicode(combo)) {
+		for (int i = 0; i < mem.Length(); i++) {
+			//Platform::DebugPrintf("Combo[%0d] = %s\n", i, mem.At(i).c_str());
+			WCHAR wszBuf[CTL_TEXT_BUF];
+			::MultiByteToWideChar(CP_UTF8, 0, mem.At(i).c_str(), -1, wszBuf,
+			                    CTL_TEXT_BUF);
+			::SendMessageW(combo, CB_ADDSTRING, 0,
+			               reinterpret_cast<LPARAM>(wszBuf));
+		}
+	} else {
+		for (int i = 0; i < mem.Length(); i++) {
+			//Platform::DebugPrintf("Combo[%0d] = %s\n", i, mem.At(i).c_str());
+			::SendMessage(combo, CB_ADDSTRING, 0,
+			              reinterpret_cast<LPARAM>(mem.At(i).c_str()));
+		}
 	}
 	if (useTop) {
 		::SendMessage(combo, CB_SETCURSEL, 0, 0);
@@ -755,9 +823,13 @@ static void FillComboFromProps(HWND combo, PropSet &props) {
 }
 
 static SString GetItemText(HWND hDlg, int id) {
-	char s[10000];
-	::GetDlgItemText(hDlg, id, s, sizeof(s));
-	return SString(s);
+	HWND wT = ::GetDlgItem(hDlg, id);
+	int len = ::GetWindowTextLength(wT);
+	SBuffer itemText(len);
+	if (len > 0) {
+		::GetDlgItemText(hDlg, id, itemText.ptr(), len + 1);
+	}
+	return SString(itemText);
 }
 
 BOOL SciTEWin::FindMessage(HWND hDlg, UINT message, WPARAM wParam) {
@@ -777,8 +849,8 @@ BOOL SciTEWin::FindMessage(HWND hDlg, UINT message, WPARAM wParam) {
 
 	case WM_INITDIALOG:
 		LocaliseDialog(hDlg);
-		::SetDlgItemText(hDlg, IDFINDWHAT, findWhat.c_str());
 		FillComboFromMemory(wFindWhat, memFinds);
+		SetDlgItemText2(hDlg, IDFINDWHAT, findWhat.c_str());
 		if (wholeWord)
 			::SendMessage(wWholeWord, BM_SETCHECK, BST_CHECKED, 0);
 		if (matchCase)
@@ -804,11 +876,10 @@ BOOL SciTEWin::FindMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		if (ControlIDOfCommand(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			wFindReplace.Destroy();
-			wFindReplace = 0;
 			return FALSE;
 		} else if ( (ControlIDOfCommand(wParam) == IDOK) ||
-					   (ControlIDOfCommand(wParam) == IDMARKALL) ){
-			findWhat = GetItemText(hDlg, IDFINDWHAT);
+		            (ControlIDOfCommand(wParam) == IDMARKALL) ) {
+			findWhat = GetDlgItemText2(hDlg, IDFINDWHAT);
 			props.Set("find.what", findWhat.c_str());
 			memFinds.Insert(findWhat.c_str());
 			wholeWord = BST_CHECKED ==
@@ -825,7 +896,6 @@ BOOL SciTEWin::FindMessage(HWND hDlg, UINT message, WPARAM wParam) {
 			              ::SendMessage(wUp, BM_GETCHECK, 0, 0);
 			::EndDialog(hDlg, IDOK);
 			wFindReplace.Destroy();
-			wFindReplace = 0;
 			if (ControlIDOfCommand(wParam) == IDMARKALL){
 				MarkAll();
 			}
@@ -852,7 +922,7 @@ BOOL SciTEWin::HandleReplaceCommand(int cmd) {
 	HWND wUnSlash = ::GetDlgItem(hwndFR, IDUNSLASH);
 
 	if ((cmd == IDOK) || (cmd == IDREPLACE) || (cmd == IDREPLACEALL) || (cmd == IDREPLACEINSEL) || (cmd == IDREPLACEINBUF)) {
-		findWhat = GetItemText(hwndFR, IDFINDWHAT);
+		findWhat = GetDlgItemText2(hwndFR, IDFINDWHAT);
 		props.Set("find.what", findWhat.c_str());
 		memFinds.Insert(findWhat.c_str());
 		wholeWord = BST_CHECKED ==
@@ -867,7 +937,7 @@ BOOL SciTEWin::HandleReplaceCommand(int cmd) {
 		          ::SendMessage(wUnSlash, BM_GETCHECK, 0, 0);
 	}
 	if ((cmd == IDREPLACE) || (cmd == IDREPLACEALL) || (cmd == IDREPLACEINSEL) || (cmd == IDREPLACEINBUF)) {
-		replaceWhat = GetItemText(hwndFR, IDREPLACEWITH);
+		replaceWhat = GetDlgItemText2(hwndFR, IDREPLACEWITH);
 		memReplaces.Insert(replaceWhat.c_str());
 	}
 
@@ -913,10 +983,10 @@ BOOL SciTEWin::ReplaceMessage(HWND hDlg, UINT message, WPARAM wParam) {
 
 	case WM_INITDIALOG:
 		LocaliseDialog(hDlg);
-		::SetDlgItemText(hDlg, IDFINDWHAT, findWhat.c_str());
 		FillComboFromMemory(wFindWhat, memFinds);
-		::SetDlgItemText(hDlg, IDREPLACEWITH, replaceWhat.c_str());
+		SetDlgItemText2(hDlg, IDFINDWHAT, findWhat.c_str());
 		FillComboFromMemory(wReplaceWith, memReplaces);
+		SetDlgItemText2(hDlg, IDREPLACEWITH, replaceWhat.c_str());
 		if (wholeWord)
 			::SendMessage(wWholeWord, BM_SETCHECK, BST_CHECKED, 0);
 		if (matchCase)
@@ -944,7 +1014,6 @@ BOOL SciTEWin::ReplaceMessage(HWND hDlg, UINT message, WPARAM wParam) {
 			UpdateStatusBar(false);
 			::EndDialog(hDlg, IDCANCEL);
 			wFindReplace.Destroy();
-			wFindReplace = 0;
 			return FALSE;
 		} else {
 			return HandleReplaceCommand(ControlIDOfCommand(wParam));
@@ -970,7 +1039,7 @@ BOOL SciTEWin::IncrementFindMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		wFindIncrement = hDlg;
 		LocaliseDialog(hDlg);
 		SetWindowLong(hDlg, GWL_STYLE, WS_TABSTOP || GetWindowLong(hDlg, GWL_STYLE));
-		::SetDlgItemText(hDlg, IDC_INCFINDTEXT, ""); //findWhat.c_str()
+		SetDlgItemText2(hDlg, IDC_INCFINDTEXT, ""); //findWhat.c_str()
 		SetFocus(hDlg);
 
 		PRectangle aRect = wFindIncrement.GetPosition();
@@ -1009,27 +1078,25 @@ BOOL SciTEWin::IncrementFindMessage(HWND hDlg, UINT message, WPARAM wParam) {
 			UpdateStatusBar(false);
 			::EndDialog(hDlg, IDCANCEL);
 			wFindIncrement.Destroy();
-			wFindIncrement = 0;
 			return FALSE;
-		} else
-		if (((ControlIDOfCommand(wParam) == IDC_INCFINDTEXT) && ((wParam >> 16) == 0x0300))
-			|| (ControlIDOfCommand(wParam) == IDC_INCFINDBTNOK)){
+		} else if (((ControlIDOfCommand(wParam) == IDC_INCFINDTEXT) && ((wParam >> 16) == 0x0300))
+			|| (ControlIDOfCommand(wParam) == IDC_INCFINDBTNOK)) {
 			SString ffLastWhat;
 			ffLastWhat = findWhat;
-			findWhat =  GetItemText(hDlg, IDC_INCFINDTEXT);
+			findWhat = GetDlgItemText2(hDlg, IDC_INCFINDTEXT);
 
-			if (ControlIDOfCommand(wParam) != IDC_INCFINDBTNOK){
+			if (ControlIDOfCommand(wParam) != IDC_INCFINDBTNOK) {
 				CharacterRange cr = GetSelection();
-				if (ffLastWhat.length()){
-					SetSelection(cr.cpMin - ffLastWhat.length(),cr.cpMin - ffLastWhat.length());
+				if (ffLastWhat.length()) {
+					SetSelection(cr.cpMin - ffLastWhat.length(), cr.cpMin - ffLastWhat.length());
 				}
 			}
 			wholeWord = false;
-			FindNext(false,false);
-			if ((!havefound)&&(ffLastWhat.length() == findWhat.length()-1) &&
-				strncmp(findWhat.c_str(),ffLastWhat.c_str(), ffLastWhat.length()) == 0 ){
+			FindNext(false, false);
+ 			if ((!havefound) &&
+				strncmp(findWhat.c_str(), ffLastWhat.c_str(), ffLastWhat.length()) == 0) {
 				findWhat = ffLastWhat;
-				::SetDlgItemText(hDlg, IDC_INCFINDTEXT, findWhat.c_str());
+				SetDlgItemText2(hDlg, IDC_INCFINDTEXT, findWhat.c_str());
 				SendMessage(wFindWhat, EM_SETSEL, ffLastWhat.length(), ffLastWhat.length());
 			}
 			return FALSE;
@@ -1060,10 +1127,19 @@ void SciTEWin::FindIncrement() {
 	fr.wFindWhatLen = static_cast<WORD>(findWhat.length() + 1);
 
 	replacing = false;
-	DoDialog(hInstance,
-		   MAKEINTRESOURCE(IDD_FIND2),
-		   MainHWND(),
-		   reinterpret_cast<DLGPROC>(FindIncrementDlg));
+	//DoDialog(hInstance,
+	//		 MAKEINTRESOURCE(IDD_FIND2),
+	//		 MainHWND(),
+	//		 reinterpret_cast<DLGPROC>(FindIncrementDlg));
+	if (IsWindowsNT()) {
+		::DialogBoxParamW(hInstance, (LPCWSTR)MAKEINTRESOURCE(IDD_FIND2),
+		                MainHWND(), reinterpret_cast<DLGPROC>(FindIncrementDlg),
+		                reinterpret_cast<LPARAM>(this));
+	} else {
+		::DialogBoxParamA(hInstance, MAKEINTRESOURCE(IDD_FIND2),
+		                MainHWND(), reinterpret_cast<DLGPROC>(FindIncrementDlg),
+		                reinterpret_cast<LPARAM>(this));
+	}
 	WindowSetFocus(wEditor);
 }
 
@@ -1084,11 +1160,19 @@ void SciTEWin::Find() {
 	fr.lpstrFindWhat = const_cast<char *>(findWhat.c_str());
 	fr.wFindWhatLen = static_cast<WORD>(findWhat.length() + 1);
 
-	wFindReplace = ::CreateDialogParam(hInstance,
-	                                   MAKEINTRESOURCE(IDD_FIND),
-	                                   MainHWND(),
-	                                   reinterpret_cast<DLGPROC>(FindDlg),
-	                                   reinterpret_cast<LPARAM>(this));
+	if (IsWindowsNT()) {
+		wFindReplace = ::CreateDialogParamW(hInstance,
+		                                    (LPCWSTR)MAKEINTRESOURCE(IDD_FIND),
+		                                    MainHWND(),
+		                                    reinterpret_cast<DLGPROC>(FindDlg),
+		                                    reinterpret_cast<LPARAM>(this));
+	} else {
+		wFindReplace = ::CreateDialogParamA(hInstance,
+		                                    MAKEINTRESOURCE(IDD_FIND),
+		                                    MainHWND(),
+		                                    reinterpret_cast<DLGPROC>(FindDlg),
+		                                    reinterpret_cast<LPARAM>(this));
+	}
 	wFindReplace.Show();
 
 	replacing = false;
@@ -1104,23 +1188,38 @@ static int __stdcall BrowseCallbackProc(HWND hwnd, UINT uMsg, LPARAM, LPARAM pDa
 	return 0;
 }
 
-BOOL SciTEWin::GrepMessage(HWND hDlg, UINT message, WPARAM wParam) {
-	HWND hFindWhat;
-	HWND hFiles;
-	HWND hDirectory;
+void SciTEWin::PerformGrep() {
+	SelectionIntoProperties();
 
+	SString findInput;
+	long flags = 0;
+	if (props.Get("find.input").length()) {
+		findInput = props.GetNewExpand("find.input");
+		flags += jobHasInput;
+	}
+
+	AddCommand(props.GetNewExpand("find.command"),
+		   props.Get("find.directory"),
+		   jobCLI, findInput, flags);
+	if (commandCurrent > 0) {
+		Execute();
+	}
+}
+
+void SciTEWin::FillCombos(HWND hDlg) {
+	FillComboFromMemory(::GetDlgItem(hDlg, IDFINDWHAT), memFinds, true);
+	FillComboFromMemory(::GetDlgItem(hDlg, IDFILES), memFiles, true);
+	FillComboFromMemory(::GetDlgItem(hDlg, IDDIRECTORY), memDirectory, true);
+}
+
+BOOL SciTEWin::GrepMessage(HWND hDlg, UINT message, WPARAM wParam) {
 	switch (message) {
 
 	case WM_INITDIALOG:
 		LocaliseDialog(hDlg);
+		FillCombos(hDlg);
 		::SetDlgItemText(hDlg, IDFINDWHAT, props.Get("find.what").c_str());
-		hFindWhat = ::GetDlgItem(hDlg, IDFINDWHAT);
-		FillComboFromMemory(hFindWhat, memFinds);
-		hFiles = ::GetDlgItem(hDlg, IDFILES);
-		FillComboFromMemory(hFiles, memFiles, true);
 		::SetDlgItemText(hDlg, IDDIRECTORY, props.Get("find.directory").c_str());
-		hDirectory = ::GetDlgItem(hDlg, IDDIRECTORY);
-		FillComboFromMemory(hDirectory, memDirectory);
 		return TRUE;
 
 	case WM_CLOSE:
@@ -1130,6 +1229,7 @@ BOOL SciTEWin::GrepMessage(HWND hDlg, UINT message, WPARAM wParam) {
 	case WM_COMMAND:
 		if (ControlIDOfCommand(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
+			wFindInFiles.Destroy();
 			return FALSE;
 
 		} else if (ControlIDOfCommand(wParam) == IDOK) {
@@ -1145,8 +1245,24 @@ BOOL SciTEWin::GrepMessage(HWND hDlg, UINT message, WPARAM wParam) {
 			props.Set("find.directory", directory.c_str());
 			memDirectory.Insert(directory.c_str());
 
-			::EndDialog(hDlg, IDOK);
-			return TRUE;
+			FillCombos(hDlg);
+
+			PerformGrep();
+			if (props.GetInt("find.in.files.close.on.find", 1)) {
+				::EndDialog(hDlg, IDOK);
+				wFindInFiles.Destroy();
+				return TRUE;
+			} else {
+				return FALSE;
+			}
+		} else if (ControlIDOfCommand(wParam) == IDDOTDOT) {
+
+			SString directory = GetItemText(hDlg, IDDIRECTORY);
+			const char *lastSlash = strrchr(directory.c_str(), pathSepChar);
+			if (lastSlash) {
+				directory.remove(lastSlash - directory.c_str(), 0);
+				::SetDlgItemText(hDlg, IDDIRECTORY, directory.c_str());
+			}
 
 		} else if (ControlIDOfCommand(wParam) == IDBROWSE) {
 
@@ -1202,28 +1318,16 @@ BOOL CALLBACK SciTEWin::GrepDlg(HWND hDlg, UINT message, WPARAM wParam, LPARAM l
 }
 
 void SciTEWin::FindInFiles() {
+	if (wFindInFiles.Created())
+		return;
 	SelectionIntoFind();
 	props.Set("find.what", findWhat.c_str());
 	char findInDir[1024];
 	GetDocumentDirectory(findInDir, sizeof(findInDir));
 	props.Set("find.directory", findInDir);
-	if (DoDialog(hInstance, "Grep", MainHWND(), reinterpret_cast<DLGPROC>(GrepDlg)) == IDOK) {
-		//Platform::DebugPrintf("asked to find %s %s %s\n", props.Get("find.what"), props.Get("find.files"), props.Get("find.directory"));
-		SelectionIntoProperties();
-
-		SString findInput;
-		long flags = 0;
-		if (props.Get("find.input").length()) {
-			findInput = props.GetNewExpand("find.input");
-			flags += jobHasInput;
-		}
-
-		AddCommand(props.GetNewExpand("find.command"),
-		           props.Get("find.directory"),
-		           jobCLI, findInput, flags);
-		if (commandCurrent > 0)
-			Execute();
-	}
+	wFindInFiles = ::CreateDialogParam(hInstance, "Grep", MainHWND(), 
+		reinterpret_cast<DLGPROC>(GrepDlg), reinterpret_cast<sptr_t>(this));
+	wFindInFiles.Show();
 }
 
 void SciTEWin::Replace() {
@@ -1241,11 +1345,19 @@ void SciTEWin::Replace() {
 	fr.wFindWhatLen = static_cast<WORD>(findWhat.length() + 1);
 	fr.wReplaceWithLen = static_cast<WORD>(replaceWhat.length() + 1);
 
-	wFindReplace = ::CreateDialogParam(hInstance,
-	                                   MAKEINTRESOURCE(IDD_REPLACE),
-	                                   MainHWND(),
-	                                   reinterpret_cast<DLGPROC>(ReplaceDlg),
-	                                   reinterpret_cast<sptr_t>(this));
+	if (IsWindowsNT()) {
+		wFindReplace = ::CreateDialogParamW(hInstance,
+		                                    (LPCWSTR)MAKEINTRESOURCE(IDD_REPLACE),
+		                                    MainHWND(),
+		                                    reinterpret_cast<DLGPROC>(ReplaceDlg),
+		                                    reinterpret_cast<sptr_t>(this));
+	} else {
+		wFindReplace = ::CreateDialogParamA(hInstance,
+		                                    MAKEINTRESOURCE(IDD_REPLACE),
+		                                    MainHWND(),
+		                                    reinterpret_cast<DLGPROC>(ReplaceDlg),
+		                                    reinterpret_cast<sptr_t>(this));
+	}
 	wFindReplace.Show();
 
 	replacing = true;
@@ -1260,7 +1372,6 @@ void SciTEWin::DestroyFindReplace() {
 	if (wFindReplace.Created()) {
 		::EndDialog(reinterpret_cast<HWND>(wFindReplace.GetID()), IDCANCEL);
 		wFindReplace.Destroy();
-		wFindReplace = 0;
 	}
 }
 
@@ -1524,7 +1635,7 @@ int SciTEWin::WindowMessageBox(Window &w, const SString &msg, int style) {
 	dialogsOnScreen++;
 	int ret = ::MessageBox(reinterpret_cast<HWND>(w.GetID()), msg.c_str(), appName, style | MB_SETFOREGROUND);
 	dialogsOnScreen--;
-	return  ret;
+	return ret;
 }
 
 BOOL SciTEWin::AboutMessage(HWND hDlg, UINT message, WPARAM wParam) {
